"""
Speech Memory Management.

Provides memory for speech operations:
- Transcription history
- Voice profiles
- Audio session cache
- Conversation transcripts
"""

import hashlib
import logging
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional

from .manager import MemoryManager

logger = logging.getLogger(__name__)


@dataclass
class TranscriptionMemory:
    """Memory of a transcription."""
    transcription_id: str
    audio_hash: str
    text: str
    language: str
    confidence: float
    provider: str
    model: str
    duration_ms: int
    word_count: int
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    words: List[Dict[str, Any]] = field(default_factory=list)
    speaker_id: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "transcription_id": self.transcription_id,
            "audio_hash": self.audio_hash,
            "text": self.text,
            "language": self.language,
            "confidence": self.confidence,
            "provider": self.provider,
            "model": self.model,
            "duration_ms": self.duration_ms,
            "word_count": self.word_count,
            "timestamp": self.timestamp,
            "words": self.words,
            "speaker_id": self.speaker_id,
            "metadata": self.metadata,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TranscriptionMemory":
        return cls(
            transcription_id=data["transcription_id"],
            audio_hash=data["audio_hash"],
            text=data["text"],
            language=data.get("language", "en"),
            confidence=data.get("confidence", 1.0),
            provider=data.get("provider", ""),
            model=data.get("model", ""),
            duration_ms=data.get("duration_ms", 0),
            word_count=data.get("word_count", len(data.get("text", "").split())),
            timestamp=data.get("timestamp", datetime.now().isoformat()),
            words=data.get("words", []),
            speaker_id=data.get("speaker_id"),
            metadata=data.get("metadata", {}),
        )


@dataclass
class SynthesisMemory:
    """Memory of a TTS synthesis."""
    synthesis_id: str
    text_hash: str
    text: str
    voice: str
    provider: str
    model: str
    audio_format: str
    duration_ms: int
    audio_size_bytes: int
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "synthesis_id": self.synthesis_id,
            "text_hash": self.text_hash,
            "text": self.text,
            "voice": self.voice,
            "provider": self.provider,
            "model": self.model,
            "audio_format": self.audio_format,
            "duration_ms": self.duration_ms,
            "audio_size_bytes": self.audio_size_bytes,
            "timestamp": self.timestamp,
            "metadata": self.metadata,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SynthesisMemory":
        return cls(
            synthesis_id=data["synthesis_id"],
            text_hash=data["text_hash"],
            text=data["text"],
            voice=data.get("voice", ""),
            provider=data.get("provider", ""),
            model=data.get("model", ""),
            audio_format=data.get("audio_format", "mp3"),
            duration_ms=data.get("duration_ms", 0),
            audio_size_bytes=data.get("audio_size_bytes", 0),
            timestamp=data.get("timestamp", datetime.now().isoformat()),
            metadata=data.get("metadata", {}),
        )


@dataclass
class VoiceProfile:
    """Profile for a voice/speaker."""
    profile_id: str
    name: str
    profile_type: str  # user, agent, system
    preferred_language: str = "en"
    speaking_rate: float = 1.0
    pitch: float = 1.0
    voice_settings: Dict[str, Any] = field(default_factory=dict)
    embedding: Optional[List[float]] = None  # Voice embedding for speaker ID
    sample_count: int = 0
    created_at: str = field(default_factory=lambda: datetime.now().isoformat())
    updated_at: str = field(default_factory=lambda: datetime.now().isoformat())
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "profile_id": self.profile_id,
            "name": self.name,
            "profile_type": self.profile_type,
            "preferred_language": self.preferred_language,
            "speaking_rate": self.speaking_rate,
            "pitch": self.pitch,
            "voice_settings": self.voice_settings,
            "embedding": self.embedding,
            "sample_count": self.sample_count,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "VoiceProfile":
        return cls(
            profile_id=data["profile_id"],
            name=data["name"],
            profile_type=data.get("profile_type", "user"),
            preferred_language=data.get("preferred_language", "en"),
            speaking_rate=data.get("speaking_rate", 1.0),
            pitch=data.get("pitch", 1.0),
            voice_settings=data.get("voice_settings", {}),
            embedding=data.get("embedding"),
            sample_count=data.get("sample_count", 0),
            created_at=data.get("created_at", datetime.now().isoformat()),
            updated_at=data.get("updated_at", datetime.now().isoformat()),
        )


@dataclass
class VoiceConversationMemory:
    """Memory of a voice conversation."""
    conversation_id: str
    participants: List[str]  # profile_ids
    turns: List[Dict[str, Any]]  # role, text, audio_id, timestamp
    total_duration_ms: int = 0
    turn_count: int = 0
    started_at: str = field(default_factory=lambda: datetime.now().isoformat())
    ended_at: Optional[str] = None
    summary: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "conversation_id": self.conversation_id,
            "participants": self.participants,
            "turns": self.turns,
            "total_duration_ms": self.total_duration_ms,
            "turn_count": self.turn_count,
            "started_at": self.started_at,
            "ended_at": self.ended_at,
            "summary": self.summary,
            "metadata": self.metadata,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "VoiceConversationMemory":
        return cls(
            conversation_id=data["conversation_id"],
            participants=data.get("participants", []),
            turns=data.get("turns", []),
            total_duration_ms=data.get("total_duration_ms", 0),
            turn_count=data.get("turn_count", 0),
            started_at=data.get("started_at", datetime.now().isoformat()),
            ended_at=data.get("ended_at"),
            summary=data.get("summary"),
            metadata=data.get("metadata", {}),
        )


@dataclass
class AudioCache:
    """Cached audio data reference."""
    cache_key: str
    audio_hash: str
    format: str
    duration_ms: int
    size_bytes: int
    storage_path: Optional[str]  # Path to cached audio file
    cached_at: str = field(default_factory=lambda: datetime.now().isoformat())
    expires_at: Optional[str] = None
    hit_count: int = 0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "cache_key": self.cache_key,
            "audio_hash": self.audio_hash,
            "format": self.format,
            "duration_ms": self.duration_ms,
            "size_bytes": self.size_bytes,
            "storage_path": self.storage_path,
            "cached_at": self.cached_at,
            "expires_at": self.expires_at,
            "hit_count": self.hit_count,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "AudioCache":
        return cls(
            cache_key=data["cache_key"],
            audio_hash=data["audio_hash"],
            format=data.get("format", "mp3"),
            duration_ms=data.get("duration_ms", 0),
            size_bytes=data.get("size_bytes", 0),
            storage_path=data.get("storage_path"),
            cached_at=data.get("cached_at", datetime.now().isoformat()),
            expires_at=data.get("expires_at"),
            hit_count=data.get("hit_count", 0),
        )
    
    @property
    def is_expired(self) -> bool:
        if not self.expires_at:
            return False
        return datetime.fromisoformat(self.expires_at) < datetime.now()


class SpeechMemoryManager:
    """
    Manages memory for speech operations.
    
    Example:
        >>> speech_memory = SpeechMemoryManager()
        >>> 
        >>> # Store transcription
        >>> speech_memory.store_transcription(audio_hash, text, language="en")
        >>> 
        >>> # Get transcription history
        >>> history = speech_memory.get_transcription_history(speaker_id="user-1")
        >>> 
        >>> # Create voice profile
        >>> profile = speech_memory.create_voice_profile("John", profile_type="user")
        >>> 
        >>> # Start voice conversation
        >>> conv = speech_memory.start_conversation(["user-1", "agent-1"])
    """
    
    def __init__(
        self,
        memory_manager: MemoryManager = None,
        audio_cache_ttl: int = 3600,  # 1 hour
        max_transcription_history: int = 500,
        max_synthesis_history: int = 500,
    ):
        self.memory = memory_manager or MemoryManager()
        self.audio_cache_ttl = audio_cache_ttl
        self.max_transcription_history = max_transcription_history
        self.max_synthesis_history = max_synthesis_history
        
        # In-memory storage
        self._transcriptions: List[TranscriptionMemory] = []
        self._syntheses: List[SynthesisMemory] = []
        self._voice_profiles: Dict[str, VoiceProfile] = {}
        self._conversations: Dict[str, VoiceConversationMemory] = {}
        self._audio_cache: Dict[str, AudioCache] = {}
    
    def _hash_audio(self, audio_data: bytes) -> str:
        """Create hash from audio data."""
        return hashlib.sha256(audio_data).hexdigest()[:16]
    
    def _hash_text(self, text: str) -> str:
        """Create hash from text."""
        return hashlib.sha256(text.encode()).hexdigest()[:16]
    
    # =========================================================================
    # Transcription History
    # =========================================================================
    
    def store_transcription(
        self,
        audio_hash: str,
        text: str,
        language: str = "en",
        confidence: float = 1.0,
        provider: str = "",
        model: str = "",
        duration_ms: int = 0,
        words: List[Dict] = None,
        speaker_id: str = None,
        metadata: Dict = None,
    ) -> TranscriptionMemory:
        """Store a transcription in memory."""
        import uuid
        
        transcription = TranscriptionMemory(
            transcription_id=f"trans-{uuid.uuid4().hex[:8]}",
            audio_hash=audio_hash,
            text=text,
            language=language,
            confidence=confidence,
            provider=provider,
            model=model,
            duration_ms=duration_ms,
            word_count=len(text.split()),
            words=words or [],
            speaker_id=speaker_id,
            metadata=metadata or {},
        )
        
        self._transcriptions.append(transcription)
        
        # Limit history
        while len(self._transcriptions) > self.max_transcription_history:
            self._transcriptions.pop(0)
        
        # Persist
        self.memory.store_long_term(
            "speech:transcriptions",
            [t.to_dict() for t in self._transcriptions[-100:]],
            priority=6,
        )
        
        return transcription
    
    def get_transcription_history(
        self,
        speaker_id: str = None,
        language: str = None,
        provider: str = None,
        last_n: int = None,
    ) -> List[TranscriptionMemory]:
        """Get transcription history."""
        if not self._transcriptions:
            data = self.memory.retrieve("speech:transcriptions", [])
            self._transcriptions = [TranscriptionMemory.from_dict(t) for t in data]
        
        history = self._transcriptions
        
        if speaker_id:
            history = [t for t in history if t.speaker_id == speaker_id]
        
        if language:
            history = [t for t in history if t.language == language]
        
        if provider:
            history = [t for t in history if t.provider == provider]
        
        if last_n:
            history = history[-last_n:]
        
        return history
    
    def search_transcriptions(
        self,
        query: str,
        top_k: int = 10,
    ) -> List[TranscriptionMemory]:
        """Search transcriptions by text content."""
        query_lower = query.lower()
        
        results = [
            t for t in self._transcriptions
            if query_lower in t.text.lower()
        ]
        
        return results[:top_k]
    
    def get_transcription_by_audio(
        self,
        audio_hash: str,
    ) -> Optional[TranscriptionMemory]:
        """Get transcription by audio hash (for caching)."""
        for t in reversed(self._transcriptions):
            if t.audio_hash == audio_hash:
                return t
        return None
    
    # =========================================================================
    # Synthesis History
    # =========================================================================
    
    def store_synthesis(
        self,
        text: str,
        voice: str,
        provider: str = "",
        model: str = "",
        audio_format: str = "mp3",
        duration_ms: int = 0,
        audio_size_bytes: int = 0,
        metadata: Dict = None,
    ) -> SynthesisMemory:
        """Store a synthesis record."""
        import uuid
        
        synthesis = SynthesisMemory(
            synthesis_id=f"synth-{uuid.uuid4().hex[:8]}",
            text_hash=self._hash_text(text),
            text=text,
            voice=voice,
            provider=provider,
            model=model,
            audio_format=audio_format,
            duration_ms=duration_ms,
            audio_size_bytes=audio_size_bytes,
            metadata=metadata or {},
        )
        
        self._syntheses.append(synthesis)
        
        # Limit history
        while len(self._syntheses) > self.max_synthesis_history:
            self._syntheses.pop(0)
        
        # Persist
        self.memory.store_long_term(
            "speech:syntheses",
            [s.to_dict() for s in self._syntheses[-100:]],
            priority=5,
        )
        
        return synthesis
    
    def get_synthesis_history(
        self,
        voice: str = None,
        provider: str = None,
        last_n: int = None,
    ) -> List[SynthesisMemory]:
        """Get synthesis history."""
        if not self._syntheses:
            data = self.memory.retrieve("speech:syntheses", [])
            self._syntheses = [SynthesisMemory.from_dict(s) for s in data]
        
        history = self._syntheses
        
        if voice:
            history = [s for s in history if s.voice == voice]
        
        if provider:
            history = [s for s in history if s.provider == provider]
        
        if last_n:
            history = history[-last_n:]
        
        return history
    
    def get_synthesis_by_text(
        self,
        text: str,
        voice: str,
    ) -> Optional[SynthesisMemory]:
        """Get synthesis by text and voice (for caching)."""
        text_hash = self._hash_text(text)
        for s in reversed(self._syntheses):
            if s.text_hash == text_hash and s.voice == voice:
                return s
        return None
    
    # =========================================================================
    # Voice Profiles
    # =========================================================================
    
    def create_voice_profile(
        self,
        name: str,
        profile_type: str = "user",
        preferred_language: str = "en",
        voice_settings: Dict = None,
    ) -> VoiceProfile:
        """Create a voice profile."""
        import uuid
        profile_id = f"voice-{uuid.uuid4().hex[:8]}"
        
        profile = VoiceProfile(
            profile_id=profile_id,
            name=name,
            profile_type=profile_type,
            preferred_language=preferred_language,
            voice_settings=voice_settings or {},
        )
        
        self._voice_profiles[profile_id] = profile
        
        self.memory.store_long_term(
            f"speech:profile:{profile_id}",
            profile.to_dict(),
            priority=7,
        )
        
        return profile
    
    def get_voice_profile(self, profile_id: str) -> Optional[VoiceProfile]:
        """Get a voice profile."""
        if profile_id in self._voice_profiles:
            return self._voice_profiles[profile_id]
        
        data = self.memory.retrieve(f"speech:profile:{profile_id}")
        if data:
            profile = VoiceProfile.from_dict(data)
            self._voice_profiles[profile_id] = profile
            return profile
        
        return None
    
    def update_voice_profile(
        self,
        profile_id: str,
        **updates,
    ) -> Optional[VoiceProfile]:
        """Update a voice profile."""
        profile = self.get_voice_profile(profile_id)
        if not profile:
            return None
        
        for key, value in updates.items():
            if hasattr(profile, key):
                setattr(profile, key, value)
        
        profile.updated_at = datetime.now().isoformat()
        
        self.memory.store_long_term(
            f"speech:profile:{profile_id}",
            profile.to_dict(),
        )
        
        return profile
    
    def set_voice_embedding(
        self,
        profile_id: str,
        embedding: List[float],
    ) -> bool:
        """Set voice embedding for speaker identification."""
        profile = self.get_voice_profile(profile_id)
        if not profile:
            return False
        
        profile.embedding = embedding
        profile.sample_count += 1
        profile.updated_at = datetime.now().isoformat()
        
        self.memory.store_long_term(
            f"speech:profile:{profile_id}",
            profile.to_dict(),
        )
        
        return True
    
    def find_speaker_by_embedding(
        self,
        embedding: List[float],
        threshold: float = 0.8,
    ) -> Optional[VoiceProfile]:
        """Find speaker by voice embedding similarity."""
        best_match = None
        best_similarity = 0.0
        
        for profile in self._voice_profiles.values():
            if profile.embedding:
                # Cosine similarity
                similarity = self._cosine_similarity(embedding, profile.embedding)
                if similarity > best_similarity and similarity >= threshold:
                    best_similarity = similarity
                    best_match = profile
        
        return best_match
    
    def _cosine_similarity(self, a: List[float], b: List[float]) -> float:
        """Calculate cosine similarity between two vectors."""
        if len(a) != len(b):
            return 0.0
        
        dot_product = sum(x * y for x, y in zip(a, b))
        norm_a = sum(x * x for x in a) ** 0.5
        norm_b = sum(x * x for x in b) ** 0.5
        
        if norm_a == 0 or norm_b == 0:
            return 0.0
        
        return dot_product / (norm_a * norm_b)
    
    # =========================================================================
    # Voice Conversations
    # =========================================================================
    
    def start_conversation(
        self,
        participants: List[str],
        metadata: Dict = None,
    ) -> VoiceConversationMemory:
        """Start a new voice conversation."""
        import uuid
        conversation_id = f"vconv-{uuid.uuid4().hex[:8]}"
        
        conversation = VoiceConversationMemory(
            conversation_id=conversation_id,
            participants=participants,
            metadata=metadata or {},
        )
        
        self._conversations[conversation_id] = conversation
        
        self.memory.store_short_term(
            f"speech:conversation:{conversation_id}",
            conversation.to_dict(),
            ttl=7200,
            priority=6,
        )
        
        return conversation
    
    def get_conversation(
        self,
        conversation_id: str,
    ) -> Optional[VoiceConversationMemory]:
        """Get a voice conversation."""
        if conversation_id in self._conversations:
            return self._conversations[conversation_id]
        
        data = self.memory.retrieve(f"speech:conversation:{conversation_id}")
        if data:
            conv = VoiceConversationMemory.from_dict(data)
            self._conversations[conversation_id] = conv
            return conv
        
        return None
    
    def add_conversation_turn(
        self,
        conversation_id: str,
        role: str,  # user, agent
        text: str,
        speaker_id: str = None,
        audio_id: str = None,
        duration_ms: int = 0,
    ) -> bool:
        """Add a turn to a conversation."""
        conv = self.get_conversation(conversation_id)
        if not conv:
            return False
        
        turn = {
            "role": role,
            "text": text,
            "speaker_id": speaker_id,
            "audio_id": audio_id,
            "duration_ms": duration_ms,
            "timestamp": datetime.now().isoformat(),
        }
        
        conv.turns.append(turn)
        conv.turn_count = len(conv.turns)
        conv.total_duration_ms += duration_ms
        
        self.memory.store_short_term(
            f"speech:conversation:{conversation_id}",
            conv.to_dict(),
            ttl=7200,
        )
        
        return True
    
    def end_conversation(
        self,
        conversation_id: str,
        summary: str = None,
    ) -> Optional[VoiceConversationMemory]:
        """End a conversation."""
        conv = self.get_conversation(conversation_id)
        if not conv:
            return None
        
        conv.ended_at = datetime.now().isoformat()
        conv.summary = summary
        
        # Move to long-term storage
        self.memory.store_long_term(
            f"speech:conversation:{conversation_id}",
            conv.to_dict(),
            priority=5,
        )
        
        return conv
    
    def get_conversation_transcript(
        self,
        conversation_id: str,
        format: str = "text",  # text, chat, markdown
    ) -> Optional[str]:
        """Get conversation transcript."""
        conv = self.get_conversation(conversation_id)
        if not conv:
            return None
        
        if format == "chat":
            return "\n".join([f"{t['role']}: {t['text']}" for t in conv.turns])
        elif format == "markdown":
            lines = []
            for t in conv.turns:
                if t["role"] == "user":
                    lines.append(f"**User:** {t['text']}")
                else:
                    lines.append(f"**Agent:** {t['text']}")
            return "\n\n".join(lines)
        else:
            return "\n".join([t["text"] for t in conv.turns])
    
    # =========================================================================
    # Audio Cache
    # =========================================================================
    
    def cache_audio(
        self,
        audio_data: bytes,
        format: str,
        duration_ms: int,
        storage_path: str = None,
        ttl: int = None,
    ) -> str:
        """Cache audio data."""
        audio_hash = self._hash_audio(audio_data)
        cache_key = f"audio:{audio_hash}"
        
        expires_at = None
        if ttl or self.audio_cache_ttl:
            ttl = ttl or self.audio_cache_ttl
            expires_at = (datetime.now() + timedelta(seconds=ttl)).isoformat()
        
        cache_entry = AudioCache(
            cache_key=cache_key,
            audio_hash=audio_hash,
            format=format,
            duration_ms=duration_ms,
            size_bytes=len(audio_data),
            storage_path=storage_path,
            expires_at=expires_at,
        )
        
        self._audio_cache[cache_key] = cache_entry
        
        self.memory.store_short_term(
            f"speech:{cache_key}",
            cache_entry.to_dict(),
            ttl=ttl or self.audio_cache_ttl,
        )
        
        return cache_key
    
    def get_audio_cache(self, audio_hash: str) -> Optional[AudioCache]:
        """Get cached audio info."""
        cache_key = f"audio:{audio_hash}"
        
        if cache_key in self._audio_cache:
            entry = self._audio_cache[cache_key]
            if not entry.is_expired:
                entry.hit_count += 1
                return entry
            else:
                del self._audio_cache[cache_key]
        
        data = self.memory.retrieve(f"speech:{cache_key}")
        if data:
            entry = AudioCache.from_dict(data)
            if not entry.is_expired:
                entry.hit_count += 1
                self._audio_cache[cache_key] = entry
                return entry
        
        return None
    
    # =========================================================================
    # Stats & Cleanup
    # =========================================================================
    
    def get_stats(self) -> Dict[str, Any]:
        """Get speech memory statistics."""
        total_transcription_duration = sum(t.duration_ms for t in self._transcriptions)
        total_synthesis_duration = sum(s.duration_ms for s in self._syntheses)
        
        return {
            "transcription_count": len(self._transcriptions),
            "total_transcription_duration_ms": total_transcription_duration,
            "synthesis_count": len(self._syntheses),
            "total_synthesis_duration_ms": total_synthesis_duration,
            "voice_profiles": len(self._voice_profiles),
            "active_conversations": len([c for c in self._conversations.values() if not c.ended_at]),
            "audio_cache_size": len(self._audio_cache),
        }
    
    def cleanup_expired_cache(self) -> int:
        """Clean up expired audio cache."""
        count = 0
        expired = [k for k, v in self._audio_cache.items() if v.is_expired]
        for key in expired:
            del self._audio_cache[key]
            count += 1
        return count


__all__ = [
    "TranscriptionMemory",
    "SynthesisMemory",
    "VoiceProfile",
    "VoiceConversationMemory",
    "AudioCache",
    "SpeechMemoryManager",
]



"""
ReAct-Style Agentic Runner.

Provides an agentic execution loop with:
- Thought/Action/Observation cycle (ReAct pattern)
- Multi-step tool calling
- Structured input/output
- Step-by-step tracing
"""

import time
import re
import logging
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING

from .types import (
    AgentInput,
    AgentOutput,
    AgentStep,
    AgentThought,
    AgentStatus,
    StepType,
)

if TYPE_CHECKING:
    from .agent import Agent

logger = logging.getLogger(__name__)


class AgentRunner:
    """
    ReAct-style agentic runner.
    
    Provides:
    - Structured input/output
    - Multi-step tool execution loop
    - Thought/Action/Observation tracing
    - Guardrail enforcement
    - Knowledge retrieval integration
    """
    
    THOUGHT_PATTERN = re.compile(r"Thought:\s*(.+?)(?=Action:|Observation:|$)", re.DOTALL | re.IGNORECASE)
    ACTION_PATTERN = re.compile(r"Action:\s*(\w+)\s*\[(.+?)\]", re.DOTALL | re.IGNORECASE)
    FINAL_ANSWER_PATTERN = re.compile(r"Final Answer:\s*(.+)", re.DOTALL | re.IGNORECASE)
    
    def __init__(
        self,
        agent: 'Agent',
        llm_manager: Optional[Any] = None,
        knowledge: Optional[Any] = None,
        guardrail_manager: Optional[Any] = None,
        guardrail_pipeline: Optional[Any] = None,
        policy_manager: Optional[Any] = None,
        monitor: Optional[Any] = None,
        tracer: Optional[Any] = None,
    ):
        self.agent = agent
        self.llm_manager = llm_manager or agent.config.get('llm') or agent.config.get('llm_manager')
        self.knowledge = knowledge or agent.config.get('knowledge')
        self.guardrail_manager = guardrail_manager or agent.config.get('guardrail_manager')
        self.guardrail_pipeline = guardrail_pipeline or agent.config.get('guardrail_pipeline')
        self.policy_manager = policy_manager or agent.config.get('policy_manager')
        self.monitor = monitor or agent.config.get('monitor')
        self.tracer = tracer or agent.config.get('tracer')
    
    def run(self, input_data: AgentInput) -> AgentOutput:
        """
        Run the agent with structured input.
        
        Args:
            input_data: Structured AgentInput
            
        Returns:
            AgentOutput with steps, thoughts, and response
        """
        from ..tracing import tracer as global_tracer
        from ..guardrails import guardrail_manager as global_guardrail_manager
        from ..tools import agent_tool_manager
        from ..context import ContextType
        
        tracer = self.tracer or global_tracer
        guardrail_mgr = self.guardrail_manager or global_guardrail_manager
        
        start_time = time.time()
        steps: List[AgentStep] = []
        thoughts: List[AgentThought] = []
        tool_results: List[Dict[str, Any]] = []
        knowledge_results: List[Dict[str, Any]] = []
        
        # Start trace
        trace_context = tracer.start_trace(f"agent.run:{self.agent.name}")
        trace_id = trace_context.trace_id if trace_context else None
        
        # Record input step
        steps.append(AgentStep(
            step_type=StepType.INPUT,
            name="user_input",
            content=input_data.prompt,
            metadata={'system_prompt': input_data.system_prompt},
        ))
        
        def finalize(output: AgentOutput) -> AgentOutput:
            output.latency_seconds = time.time() - start_time
            output.trace_id = trace_id
            output.steps = steps
            output.thoughts = thoughts
            if trace_context:
                status = "OK" if output.is_success else "ERROR"
                tracer.end_span(trace_context, status=status)
            return output
        
        try:
            # Input guardrails
            step_start = time.time()
            guardrail_report = {}
            if self.guardrail_pipeline:
                guardrail_report = self.guardrail_pipeline.execute(input_data.prompt, context=input_data.context)
            elif guardrail_mgr:
                guardrail_report = guardrail_mgr.enforce_guardrails(input_data.prompt, fail_fast=True)
            
            steps.append(AgentStep(
                step_type=StepType.GUARDRAIL,
                name="input_guardrails",
                content=guardrail_report,
                duration_ms=(time.time() - step_start) * 1000,
            ))
            
            if not guardrail_report.get('is_valid', True):
                return finalize(AgentOutput(
                    status=AgentStatus.BLOCKED,
                    guardrail_report=guardrail_report,
                    error="Input blocked by guardrails",
                ))
            
            # Knowledge retrieval
            if self.knowledge:
                step_start = time.time()
                query = input_data.knowledge_query or input_data.prompt
                knowledge_results = self.knowledge.retrieve(query)
                steps.append(AgentStep(
                    step_type=StepType.KNOWLEDGE,
                    name="knowledge_retrieval",
                    content=knowledge_results[:5] if knowledge_results else [],
                    duration_ms=(time.time() - step_start) * 1000,
                    metadata={'query': query},
                ))
                
                if knowledge_results:
                    self.agent.context_manager.add_context(
                        f"Retrieved knowledge: {str(knowledge_results[:3])[:500]}",
                        importance=0.7,
                        context_type=ContextType.KNOWLEDGE,
                    )
            
            # Bind tools
            if input_data.tools:
                agent_tool_manager.bind_tools(self.agent, input_data.tools)
            
            # Build system prompt
            system_parts = []
            if input_data.system_prompt:
                system_parts.append(input_data.system_prompt)
            else:
                system_parts.append(f"You are {self.agent.name}, an AI assistant with role: {self.agent.role}.")
            
            if input_data.tools:
                tool_schemas = agent_tool_manager.get_all_schemas(self.agent)
                tool_desc = "\n".join(
                    f"- {t['name']}: {t.get('description', 'No description')}"
                    for t in tool_schemas
                )
                system_parts.append(f"\nAvailable tools:\n{tool_desc}")
                system_parts.append("""
When you need to use a tool, format your response as:
Thought: [your reasoning]
Action: tool_name[input parameters as JSON]

After receiving the observation, continue reasoning or provide:
Final Answer: [your final response to the user]
""")
            
            system_prompt = "\n".join(system_parts)
            
            # Build initial prompt
            context_summary = self.agent.context_manager.get_context_summary()
            prompt_parts = [system_prompt]
            if context_summary and context_summary != "No context available.":
                prompt_parts.append(f"\nContext:\n{context_summary}")
            if knowledge_results:
                knowledge_text = "\n".join(f"- {str(k)[:200]}" for k in knowledge_results[:3])
                prompt_parts.append(f"\nKnowledge:\n{knowledge_text}")
            prompt_parts.append(f"\nUser: {input_data.prompt}")
            
            current_prompt = "\n".join(prompt_parts)
            
            # Agentic loop
            iteration = 0
            final_response = None
            
            while iteration < input_data.max_iterations:
                iteration += 1
                
                # LLM call
                step_start = time.time()
                llm_response = self.llm_manager.generate(
                    current_prompt,
                    temperature=input_data.temperature,
                )
                
                if llm_response is None:
                    return finalize(AgentOutput(
                        status=AgentStatus.ERROR,
                        error="LLM generation failed",
                    ))
                
                steps.append(AgentStep(
                    step_type=StepType.LLM_CALL,
                    name=f"llm_call_{iteration}",
                    content=llm_response,
                    duration_ms=(time.time() - step_start) * 1000,
                    metadata={'iteration': iteration},
                ))
                
                # Parse response
                final_match = self.FINAL_ANSWER_PATTERN.search(llm_response)
                if final_match:
                    final_response = final_match.group(1).strip()
                    thoughts.append(AgentThought(
                        thought="Providing final answer",
                        observation=final_response,
                    ))
                    break
                
                # Parse thought
                thought_match = self.THOUGHT_PATTERN.search(llm_response)
                thought_text = thought_match.group(1).strip() if thought_match else ""
                
                # Parse action
                action_match = self.ACTION_PATTERN.search(llm_response)
                if action_match:
                    action_name = action_match.group(1).strip()
                    action_input_str = action_match.group(2).strip()
                    
                    # Parse action input
                    import json
                    try:
                        action_input = json.loads(action_input_str)
                    except json.JSONDecodeError:
                        action_input = {'input': action_input_str}
                    
                    # Check policy
                    if self.policy_manager:
                        policy_result = self.policy_manager.evaluate_policies(
                            self.agent.id,
                            action=action_name,
                            resource=action_name,
                            context={'agent_id': self.agent.id},
                        )
                        if not policy_result.get('allowed', True):
                            observation = f"Tool {action_name} blocked by policy: {policy_result.get('reasons')}"
                            thoughts.append(AgentThought(
                                thought=thought_text,
                                action=action_name,
                                action_input=action_input,
                                observation=observation,
                            ))
                            current_prompt += f"\n{llm_response}\nObservation: {observation}"
                            continue
                    
                    # Execute tool
                    step_start = time.time()
                    
                    # Merge with predefined tool inputs
                    merged_input = {**(input_data.tool_inputs or {}).get(action_name, {}), **action_input}
                    
                    tool_result = agent_tool_manager.execute_tool(self.agent, action_name, **merged_input)
                    tool_results.append(tool_result.to_dict())
                    
                    observation = str(tool_result.data) if tool_result.is_success else f"Error: {tool_result.error}"
                    
                    steps.append(AgentStep(
                        step_type=StepType.TOOL_CALL,
                        name=f"tool_{action_name}",
                        content={'action': action_name, 'input': merged_input},
                        duration_ms=(time.time() - step_start) * 1000,
                    ))
                    steps.append(AgentStep(
                        step_type=StepType.TOOL_RESULT,
                        name=f"tool_result_{action_name}",
                        content=observation[:1000],
                        duration_ms=0,
                    ))
                    
                    thoughts.append(AgentThought(
                        thought=thought_text,
                        action=action_name,
                        action_input=merged_input,
                        observation=observation[:500],
                    ))
                    
                    # Add to context
                    self.agent.context_manager.add_context(
                        f"Tool {action_name} result: {observation[:300]}",
                        importance=0.6,
                        context_type=ContextType.TOOL_RESULT,
                    )
                    
                    # Continue loop
                    current_prompt += f"\n{llm_response}\nObservation: {observation}"
                
                else:
                    # No action found, treat as final response
                    final_response = llm_response
                    thoughts.append(AgentThought(
                        thought=thought_text or "Generating response",
                        observation=llm_response[:500],
                    ))
                    break
            
            if final_response is None:
                final_response = "Max iterations reached without final answer."
            
            # Output guardrails
            step_start = time.time()
            output_report = {}
            if self.guardrail_pipeline:
                output_report = self.guardrail_pipeline.execute(final_response, context=input_data.context)
            elif guardrail_mgr:
                output_report = guardrail_mgr.enforce_guardrails(final_response, fail_fast=True)
            
            steps.append(AgentStep(
                step_type=StepType.GUARDRAIL,
                name="output_guardrails",
                content=output_report,
                duration_ms=(time.time() - step_start) * 1000,
            ))
            
            if not output_report.get('is_valid', True):
                return finalize(AgentOutput(
                    status=AgentStatus.BLOCKED,
                    guardrail_report=output_report,
                    error="Output blocked by guardrails",
                ))
            
            # Record output step
            steps.append(AgentStep(
                step_type=StepType.OUTPUT,
                name="final_output",
                content=final_response,
            ))
            
            # Add to memory
            from ..context import ContextType
            self.agent.context_manager.add_context(
                input_data.prompt,
                importance=0.5,
                context_type=ContextType.USER,
            )
            self.agent.context_manager.add_context(
                final_response,
                importance=0.6,
                context_type=ContextType.ASSISTANT,
            )
            
            # Monitor
            if self.monitor:
                self.monitor.record_metric('agent.execution_seconds', time.time() - start_time)
                self.monitor.log_event('agent.run_complete', {
                    'agent_id': self.agent.id,
                    'iterations': iteration,
                    'tools_used': [t.get('tool_name') for t in tool_results],
                })
            
            return finalize(AgentOutput(
                status=AgentStatus.SUCCESS,
                response=final_response,
                tool_results=tool_results,
                knowledge_results=knowledge_results,
                guardrail_report=guardrail_report,
            ))
        
        except Exception as e:
            logger.exception("AgentRunner error")
            steps.append(AgentStep(
                step_type=StepType.ERROR,
                name="error",
                content=str(e),
            ))
            return finalize(AgentOutput(
                status=AgentStatus.ERROR,
                error=str(e),
            ))


__all__ = ['AgentRunner']



"""
Enterprise Data Validator Module.

Schema validation, data sanitization, transformation,
and validation rules engine.

Example:
    # Create validator
    validator = create_validator()
    
    # Define schema
    schema = validator.schema({
        "name": {"type": "string", "required": True, "min_length": 1},
        "email": {"type": "string", "pattern": r"^[\w.-]+@[\w.-]+\.\w+$"},
        "age": {"type": "integer", "min": 0, "max": 150},
    })
    
    # Validate data
    result = await validator.validate(
        data={"name": "John", "email": "john@example.com", "age": 30},
        schema=schema,
    )
    
    if result.is_valid:
        print("Data is valid")
"""

from __future__ import annotations

import asyncio
import logging
import re
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, date
from decimal import Decimal
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class ValidationError(Exception):
    """Validation error."""
    pass


class SchemaError(Exception):
    """Schema error."""
    pass


class ValidationType(str, Enum):
    """Validation type."""
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    NUMBER = "number"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"
    DATE = "date"
    DATETIME = "datetime"
    EMAIL = "email"
    URL = "url"
    UUID = "uuid"
    ANY = "any"


class SanitizeMode(str, Enum):
    """Sanitize mode."""
    NONE = "none"
    TRIM = "trim"
    LOWER = "lower"
    UPPER = "upper"
    STRIP_HTML = "strip_html"
    ESCAPE_HTML = "escape_html"


@dataclass
class FieldError:
    """Field validation error."""
    field: str = ""
    message: str = ""
    value: Any = None
    rule: str = ""
    path: str = ""


@dataclass
class ValidationResult:
    """Validation result."""
    is_valid: bool = True
    errors: List[FieldError] = field(default_factory=list)
    data: Any = None
    sanitized_data: Any = None
    duration: float = 0.0
    
    def add_error(
        self,
        field: str,
        message: str,
        value: Any = None,
        rule: str = "",
        path: str = "",
    ) -> None:
        """Add error."""
        self.errors.append(FieldError(
            field=field,
            message=message,
            value=value,
            rule=rule,
            path=path or field,
        ))
        self.is_valid = False
    
    def merge(self, other: ValidationResult) -> None:
        """Merge another result."""
        self.errors.extend(other.errors)
        if not other.is_valid:
            self.is_valid = False


@dataclass
class FieldSchema:
    """Field schema."""
    type: ValidationType = ValidationType.ANY
    required: bool = False
    nullable: bool = True
    default: Any = None
    
    # String constraints
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: Optional[str] = None
    
    # Numeric constraints
    min: Optional[float] = None
    max: Optional[float] = None
    
    # Array constraints
    min_items: Optional[int] = None
    max_items: Optional[int] = None
    items: Optional[Dict[str, Any]] = None
    unique: bool = False
    
    # Object constraints
    properties: Optional[Dict[str, Any]] = None
    
    # Enum
    enum: Optional[List[Any]] = None
    
    # Custom
    validator: Optional[Callable] = None
    sanitizer: Optional[Callable] = None
    sanitize_mode: SanitizeMode = SanitizeMode.NONE
    
    # Metadata
    description: str = ""
    example: Any = None


@dataclass
class Schema:
    """Validation schema."""
    fields: Dict[str, FieldSchema] = field(default_factory=dict)
    strict: bool = False  # Fail on unknown fields
    allow_extra: bool = True
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Schema:
        """Create from dictionary."""
        fields = {}
        
        for name, config in data.items():
            if isinstance(config, dict):
                field_type = config.get("type", "any")
                if isinstance(field_type, str):
                    field_type = ValidationType(field_type)
                
                fields[name] = FieldSchema(
                    type=field_type,
                    required=config.get("required", False),
                    nullable=config.get("nullable", True),
                    default=config.get("default"),
                    min_length=config.get("min_length"),
                    max_length=config.get("max_length"),
                    pattern=config.get("pattern"),
                    min=config.get("min"),
                    max=config.get("max"),
                    min_items=config.get("min_items"),
                    max_items=config.get("max_items"),
                    items=config.get("items"),
                    unique=config.get("unique", False),
                    properties=config.get("properties"),
                    enum=config.get("enum"),
                    description=config.get("description", ""),
                    example=config.get("example"),
                )
        
        return cls(fields=fields)


# Built-in validators
class BuiltinValidators:
    """Built-in validators."""
    
    EMAIL_PATTERN = re.compile(r'^[\w\.-]+@[\w\.-]+\.\w+$')
    URL_PATTERN = re.compile(
        r'^https?://'
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|'
        r'localhost|'
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
        r'(?::\d+)?'
        r'(?:/?|[/?]\S+)$',
        re.IGNORECASE
    )
    UUID_PATTERN = re.compile(
        r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$',
        re.IGNORECASE
    )
    
    @classmethod
    def is_email(cls, value: str) -> bool:
        """Check if email."""
        return bool(cls.EMAIL_PATTERN.match(value))
    
    @classmethod
    def is_url(cls, value: str) -> bool:
        """Check if URL."""
        return bool(cls.URL_PATTERN.match(value))
    
    @classmethod
    def is_uuid(cls, value: str) -> bool:
        """Check if UUID."""
        return bool(cls.UUID_PATTERN.match(value))


# Sanitizers
class Sanitizers:
    """Built-in sanitizers."""
    
    HTML_TAGS = re.compile(r'<[^>]+>')
    
    @classmethod
    def trim(cls, value: str) -> str:
        """Trim whitespace."""
        return value.strip() if isinstance(value, str) else value
    
    @classmethod
    def lower(cls, value: str) -> str:
        """Convert to lowercase."""
        return value.lower() if isinstance(value, str) else value
    
    @classmethod
    def upper(cls, value: str) -> str:
        """Convert to uppercase."""
        return value.upper() if isinstance(value, str) else value
    
    @classmethod
    def strip_html(cls, value: str) -> str:
        """Strip HTML tags."""
        if isinstance(value, str):
            return cls.HTML_TAGS.sub('', value)
        return value
    
    @classmethod
    def escape_html(cls, value: str) -> str:
        """Escape HTML entities."""
        if isinstance(value, str):
            return (
                value
                .replace('&', '&amp;')
                .replace('<', '&lt;')
                .replace('>', '&gt;')
                .replace('"', '&quot;')
                .replace("'", '&#x27;')
            )
        return value
    
    @classmethod
    def apply(cls, value: Any, mode: SanitizeMode) -> Any:
        """Apply sanitization."""
        if mode == SanitizeMode.TRIM:
            return cls.trim(value)
        elif mode == SanitizeMode.LOWER:
            return cls.lower(value)
        elif mode == SanitizeMode.UPPER:
            return cls.upper(value)
        elif mode == SanitizeMode.STRIP_HTML:
            return cls.strip_html(value)
        elif mode == SanitizeMode.ESCAPE_HTML:
            return cls.escape_html(value)
        return value


# Type coercer
class TypeCoercer:
    """Type coercion."""
    
    @classmethod
    def coerce(cls, value: Any, target_type: ValidationType) -> Any:
        """Coerce value to target type."""
        if value is None:
            return None
        
        try:
            if target_type == ValidationType.STRING:
                return str(value)
            
            elif target_type == ValidationType.INTEGER:
                if isinstance(value, str):
                    return int(float(value))
                return int(value)
            
            elif target_type == ValidationType.FLOAT:
                return float(value)
            
            elif target_type == ValidationType.NUMBER:
                if isinstance(value, (int, float, Decimal)):
                    return value
                return float(value)
            
            elif target_type == ValidationType.BOOLEAN:
                if isinstance(value, bool):
                    return value
                if isinstance(value, str):
                    return value.lower() in ('true', '1', 'yes', 'on')
                return bool(value)
            
            elif target_type == ValidationType.ARRAY:
                if isinstance(value, (list, tuple)):
                    return list(value)
                return [value]
            
            elif target_type == ValidationType.DATE:
                if isinstance(value, date):
                    return value
                if isinstance(value, str):
                    return datetime.fromisoformat(value).date()
                return value
            
            elif target_type == ValidationType.DATETIME:
                if isinstance(value, datetime):
                    return value
                if isinstance(value, str):
                    return datetime.fromisoformat(value)
                return value
            
        except (ValueError, TypeError):
            pass
        
        return value


# Field validator
class FieldValidator:
    """Field validator."""
    
    def validate(
        self,
        value: Any,
        field_schema: FieldSchema,
        field_name: str,
        path: str = "",
    ) -> ValidationResult:
        """Validate field value."""
        result = ValidationResult()
        path = path or field_name
        
        # Check required
        if value is None:
            if field_schema.required:
                result.add_error(field_name, "Field is required", None, "required", path)
                return result
            elif not field_schema.nullable:
                result.add_error(field_name, "Field cannot be null", None, "nullable", path)
                return result
            return result
        
        # Type validation
        type_valid = self._validate_type(value, field_schema.type, field_name, path, result)
        
        if not type_valid:
            return result
        
        # String validations
        if field_schema.type == ValidationType.STRING:
            self._validate_string(value, field_schema, field_name, path, result)
        
        # Numeric validations
        elif field_schema.type in (ValidationType.INTEGER, ValidationType.FLOAT, ValidationType.NUMBER):
            self._validate_number(value, field_schema, field_name, path, result)
        
        # Array validations
        elif field_schema.type == ValidationType.ARRAY:
            self._validate_array(value, field_schema, field_name, path, result)
        
        # Enum validation
        if field_schema.enum and value not in field_schema.enum:
            result.add_error(
                field_name,
                f"Value must be one of: {field_schema.enum}",
                value,
                "enum",
                path,
            )
        
        # Custom validator
        if field_schema.validator and result.is_valid:
            try:
                custom_result = field_schema.validator(value)
                if custom_result is False:
                    result.add_error(field_name, "Custom validation failed", value, "custom", path)
                elif isinstance(custom_result, str):
                    result.add_error(field_name, custom_result, value, "custom", path)
            except Exception as e:
                result.add_error(field_name, str(e), value, "custom", path)
        
        return result
    
    def _validate_type(
        self,
        value: Any,
        expected_type: ValidationType,
        field_name: str,
        path: str,
        result: ValidationResult,
    ) -> bool:
        """Validate type."""
        if expected_type == ValidationType.ANY:
            return True
        
        type_checks = {
            ValidationType.STRING: lambda v: isinstance(v, str),
            ValidationType.INTEGER: lambda v: isinstance(v, int) and not isinstance(v, bool),
            ValidationType.FLOAT: lambda v: isinstance(v, (int, float)) and not isinstance(v, bool),
            ValidationType.NUMBER: lambda v: isinstance(v, (int, float, Decimal)) and not isinstance(v, bool),
            ValidationType.BOOLEAN: lambda v: isinstance(v, bool),
            ValidationType.ARRAY: lambda v: isinstance(v, (list, tuple)),
            ValidationType.OBJECT: lambda v: isinstance(v, dict),
            ValidationType.EMAIL: lambda v: isinstance(v, str) and BuiltinValidators.is_email(v),
            ValidationType.URL: lambda v: isinstance(v, str) and BuiltinValidators.is_url(v),
            ValidationType.UUID: lambda v: isinstance(v, str) and BuiltinValidators.is_uuid(v),
            ValidationType.DATE: lambda v: isinstance(v, (date, str)),
            ValidationType.DATETIME: lambda v: isinstance(v, (datetime, str)),
        }
        
        checker = type_checks.get(expected_type)
        
        if checker and not checker(value):
            result.add_error(
                field_name,
                f"Expected type {expected_type.value}, got {type(value).__name__}",
                value,
                "type",
                path,
            )
            return False
        
        return True
    
    def _validate_string(
        self,
        value: str,
        schema: FieldSchema,
        field_name: str,
        path: str,
        result: ValidationResult,
    ) -> None:
        """Validate string."""
        if schema.min_length is not None and len(value) < schema.min_length:
            result.add_error(
                field_name,
                f"Length must be at least {schema.min_length}",
                value,
                "min_length",
                path,
            )
        
        if schema.max_length is not None and len(value) > schema.max_length:
            result.add_error(
                field_name,
                f"Length must be at most {schema.max_length}",
                value,
                "max_length",
                path,
            )
        
        if schema.pattern:
            if not re.match(schema.pattern, value):
                result.add_error(
                    field_name,
                    f"Value does not match pattern: {schema.pattern}",
                    value,
                    "pattern",
                    path,
                )
    
    def _validate_number(
        self,
        value: Union[int, float],
        schema: FieldSchema,
        field_name: str,
        path: str,
        result: ValidationResult,
    ) -> None:
        """Validate number."""
        if schema.min is not None and value < schema.min:
            result.add_error(
                field_name,
                f"Value must be at least {schema.min}",
                value,
                "min",
                path,
            )
        
        if schema.max is not None and value > schema.max:
            result.add_error(
                field_name,
                f"Value must be at most {schema.max}",
                value,
                "max",
                path,
            )
    
    def _validate_array(
        self,
        value: List[Any],
        schema: FieldSchema,
        field_name: str,
        path: str,
        result: ValidationResult,
    ) -> None:
        """Validate array."""
        if schema.min_items is not None and len(value) < schema.min_items:
            result.add_error(
                field_name,
                f"Array must have at least {schema.min_items} items",
                value,
                "min_items",
                path,
            )
        
        if schema.max_items is not None and len(value) > schema.max_items:
            result.add_error(
                field_name,
                f"Array must have at most {schema.max_items} items",
                value,
                "max_items",
                path,
            )
        
        if schema.unique:
            seen = set()
            for i, item in enumerate(value):
                key = str(item) if not isinstance(item, (str, int, float, bool)) else item
                if key in seen:
                    result.add_error(
                        field_name,
                        f"Duplicate item at index {i}",
                        item,
                        "unique",
                        f"{path}[{i}]",
                    )
                seen.add(key)


# Data validator
class DataValidator:
    """Data validator."""
    
    def __init__(self):
        self._field_validator = FieldValidator()
        self._schemas: Dict[str, Schema] = {}
    
    def schema(self, definition: Dict[str, Any]) -> Schema:
        """Create schema from definition."""
        return Schema.from_dict(definition)
    
    def register_schema(self, name: str, schema: Schema) -> None:
        """Register named schema."""
        self._schemas[name] = schema
    
    def get_schema(self, name: str) -> Optional[Schema]:
        """Get registered schema."""
        return self._schemas.get(name)
    
    async def validate(
        self,
        data: Dict[str, Any],
        schema: Union[str, Schema, Dict[str, Any]],
        coerce: bool = False,
        sanitize: bool = True,
    ) -> ValidationResult:
        """Validate data against schema."""
        start = time.time()
        result = ValidationResult(data=data)
        
        # Resolve schema
        if isinstance(schema, str):
            schema = self._schemas.get(schema)
            if not schema:
                result.add_error("_schema", f"Schema not found: {schema}")
                return result
        elif isinstance(schema, dict):
            schema = Schema.from_dict(schema)
        
        sanitized = {}
        
        # Validate each field
        for field_name, field_schema in schema.fields.items():
            value = data.get(field_name)
            
            # Apply default
            if value is None and field_schema.default is not None:
                value = field_schema.default
            
            # Coerce type
            if coerce and value is not None:
                value = TypeCoercer.coerce(value, field_schema.type)
            
            # Sanitize
            if sanitize and value is not None:
                if field_schema.sanitizer:
                    value = field_schema.sanitizer(value)
                elif field_schema.sanitize_mode != SanitizeMode.NONE:
                    value = Sanitizers.apply(value, field_schema.sanitize_mode)
            
            # Validate
            field_result = self._field_validator.validate(
                value, field_schema, field_name
            )
            
            result.merge(field_result)
            sanitized[field_name] = value
        
        # Check for unknown fields
        if schema.strict:
            for key in data:
                if key not in schema.fields:
                    result.add_error(key, "Unknown field", data[key], "strict")
        
        # Copy extra fields
        if schema.allow_extra:
            for key, value in data.items():
                if key not in sanitized:
                    sanitized[key] = value
        
        result.sanitized_data = sanitized
        result.duration = time.time() - start
        
        return result
    
    async def validate_many(
        self,
        items: List[Dict[str, Any]],
        schema: Union[str, Schema, Dict[str, Any]],
        **kwargs,
    ) -> List[ValidationResult]:
        """Validate multiple items."""
        results = []
        
        for item in items:
            result = await self.validate(item, schema, **kwargs)
            results.append(result)
        
        return results
    
    def validate_sync(
        self,
        data: Dict[str, Any],
        schema: Union[str, Schema, Dict[str, Any]],
        **kwargs,
    ) -> ValidationResult:
        """Synchronous validation."""
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(self.validate(data, schema, **kwargs))
        finally:
            loop.close()


# Validation decorator
def validated(
    schema: Union[str, Schema, Dict[str, Any]],
    validator: Optional[DataValidator] = None,
    arg_name: str = "data",
):
    """Decorator for validating function arguments."""
    def decorator(func: Callable) -> Callable:
        async def async_wrapper(*args, **kwargs):
            val = validator or DataValidator()
            data = kwargs.get(arg_name) or (args[0] if args else {})
            
            result = await val.validate(data, schema)
            
            if not result.is_valid:
                errors = "; ".join(f"{e.field}: {e.message}" for e in result.errors)
                raise ValidationError(f"Validation failed: {errors}")
            
            # Replace with sanitized data
            if arg_name in kwargs:
                kwargs[arg_name] = result.sanitized_data
            
            return await func(*args, **kwargs)
        
        def sync_wrapper(*args, **kwargs):
            val = validator or DataValidator()
            data = kwargs.get(arg_name) or (args[0] if args else {})
            
            result = val.validate_sync(data, schema)
            
            if not result.is_valid:
                errors = "; ".join(f"{e.field}: {e.message}" for e in result.errors)
                raise ValidationError(f"Validation failed: {errors}")
            
            if arg_name in kwargs:
                kwargs[arg_name] = result.sanitized_data
            
            return func(*args, **kwargs)
        
        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper
    
    return decorator


# Factory functions
def create_validator() -> DataValidator:
    """Create data validator."""
    return DataValidator()


def create_schema(definition: Dict[str, Any]) -> Schema:
    """Create validation schema."""
    return Schema.from_dict(definition)


def create_field_schema(
    type: Union[str, ValidationType] = ValidationType.ANY,
    **kwargs,
) -> FieldSchema:
    """Create field schema."""
    if isinstance(type, str):
        type = ValidationType(type)
    return FieldSchema(type=type, **kwargs)


__all__ = [
    # Exceptions
    "ValidationError",
    "SchemaError",
    # Enums
    "ValidationType",
    "SanitizeMode",
    # Data classes
    "FieldError",
    "ValidationResult",
    "FieldSchema",
    "Schema",
    # Validators
    "BuiltinValidators",
    "Sanitizers",
    "TypeCoercer",
    "FieldValidator",
    "DataValidator",
    # Decorators
    "validated",
    # Factory functions
    "create_validator",
    "create_schema",
    "create_field_schema",
]



"""
Enterprise Translation Service Module.

Internationalization (i18n), locale management,
pluralization, and translation management.

Example:
    # Create translation service
    i18n = create_translation_service()
    
    # Add translations
    await i18n.add_translations("en", {
        "greeting": "Hello, {name}!",
        "items": "{count} item|{count} items",
    })
    
    # Translate
    text = await i18n.translate("greeting", locale="en", name="John")
    # "Hello, John!"
    
    # Pluralization
    text = await i18n.translate("items", locale="en", count=5)
    # "5 items"
    
    # Change locale
    await i18n.set_locale("es")
"""

from __future__ import annotations

import asyncio
import json
import logging
import re
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class TranslationError(Exception):
    """Translation error."""
    pass


class LocaleNotFoundError(TranslationError):
    """Locale not found."""
    pass


class TranslationNotFoundError(TranslationError):
    """Translation not found."""
    pass


class PluralRule(str, Enum):
    """Plural rule."""
    ZERO = "zero"
    ONE = "one"
    TWO = "two"
    FEW = "few"
    MANY = "many"
    OTHER = "other"


@dataclass
class Locale:
    """Locale definition."""
    code: str = ""  # e.g., "en", "en-US"
    name: str = ""
    native_name: str = ""
    language: str = ""
    region: str = ""
    direction: str = "ltr"  # ltr or rtl
    date_format: str = "%Y-%m-%d"
    time_format: str = "%H:%M:%S"
    datetime_format: str = "%Y-%m-%d %H:%M:%S"
    number_format: Dict[str, str] = field(default_factory=lambda: {
        "decimal": ".",
        "thousands": ",",
    })
    currency: str = "USD"
    currency_format: str = "${amount}"
    plural_rule: str = "n != 1"  # Expression for other form


@dataclass
class Translation:
    """Translation entry."""
    key: str = ""
    value: str = ""
    locale: str = ""
    context: str = ""
    plurals: Dict[str, str] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    updated_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class TranslationBundle:
    """Translation bundle."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    locale: str = ""
    namespace: str = "default"
    translations: Dict[str, Translation] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class TranslationStats:
    """Translation statistics."""
    total_translations: int = 0
    total_locales: int = 0
    missing_translations: int = 0
    by_locale: Dict[str, int] = field(default_factory=dict)


# Translation store
class TranslationStore(ABC):
    """Translation storage."""
    
    @abstractmethod
    async def save_translation(
        self,
        locale: str,
        key: str,
        translation: Translation,
    ) -> None:
        """Save translation."""
        pass
    
    @abstractmethod
    async def get_translation(
        self,
        locale: str,
        key: str,
    ) -> Optional[Translation]:
        """Get translation."""
        pass
    
    @abstractmethod
    async def get_all(
        self,
        locale: str,
    ) -> Dict[str, Translation]:
        """Get all translations for locale."""
        pass
    
    @abstractmethod
    async def list_locales(self) -> List[str]:
        """List available locales."""
        pass


class InMemoryTranslationStore(TranslationStore):
    """In-memory translation store."""
    
    def __init__(self):
        self._translations: Dict[str, Dict[str, Translation]] = {}
    
    async def save_translation(
        self,
        locale: str,
        key: str,
        translation: Translation,
    ) -> None:
        if locale not in self._translations:
            self._translations[locale] = {}
        self._translations[locale][key] = translation
    
    async def get_translation(
        self,
        locale: str,
        key: str,
    ) -> Optional[Translation]:
        return self._translations.get(locale, {}).get(key)
    
    async def get_all(
        self,
        locale: str,
    ) -> Dict[str, Translation]:
        return self._translations.get(locale, {})
    
    async def list_locales(self) -> List[str]:
        return list(self._translations.keys())


# Locale store
class LocaleStore(ABC):
    """Locale storage."""
    
    @abstractmethod
    async def save(self, locale: Locale) -> None:
        """Save locale."""
        pass
    
    @abstractmethod
    async def get(self, code: str) -> Optional[Locale]:
        """Get locale."""
        pass
    
    @abstractmethod
    async def list(self) -> List[Locale]:
        """List locales."""
        pass


class InMemoryLocaleStore(LocaleStore):
    """In-memory locale store."""
    
    def __init__(self):
        self._locales: Dict[str, Locale] = {}
        self._init_defaults()
    
    def _init_defaults(self) -> None:
        """Initialize default locales."""
        defaults = [
            Locale(
                code="en",
                name="English",
                native_name="English",
                language="en",
                direction="ltr",
            ),
            Locale(
                code="es",
                name="Spanish",
                native_name="Español",
                language="es",
                direction="ltr",
            ),
            Locale(
                code="fr",
                name="French",
                native_name="Français",
                language="fr",
                direction="ltr",
            ),
            Locale(
                code="de",
                name="German",
                native_name="Deutsch",
                language="de",
                direction="ltr",
            ),
            Locale(
                code="zh",
                name="Chinese",
                native_name="中文",
                language="zh",
                direction="ltr",
            ),
            Locale(
                code="ja",
                name="Japanese",
                native_name="日本語",
                language="ja",
                direction="ltr",
            ),
            Locale(
                code="ar",
                name="Arabic",
                native_name="العربية",
                language="ar",
                direction="rtl",
            ),
        ]
        for locale in defaults:
            self._locales[locale.code] = locale
    
    async def save(self, locale: Locale) -> None:
        self._locales[locale.code] = locale
    
    async def get(self, code: str) -> Optional[Locale]:
        return self._locales.get(code)
    
    async def list(self) -> List[Locale]:
        return list(self._locales.values())


# Plural rules
class PluralRules:
    """Plural rules for different languages."""
    
    RULES = {
        "en": lambda n: "one" if n == 1 else "other",
        "es": lambda n: "one" if n == 1 else "other",
        "fr": lambda n: "one" if n in (0, 1) else "other",
        "de": lambda n: "one" if n == 1 else "other",
        "zh": lambda n: "other",  # No plural forms
        "ja": lambda n: "other",  # No plural forms
        "ar": lambda n: (
            "zero" if n == 0 else
            "one" if n == 1 else
            "two" if n == 2 else
            "few" if 3 <= n % 100 <= 10 else
            "many" if 11 <= n % 100 <= 99 else
            "other"
        ),
        "ru": lambda n: (
            "one" if n % 10 == 1 and n % 100 != 11 else
            "few" if 2 <= n % 10 <= 4 and (n % 100 < 10 or n % 100 >= 20) else
            "many"
        ),
    }
    
    @classmethod
    def get_plural_form(cls, locale: str, count: int) -> str:
        """Get plural form for count."""
        lang = locale.split("-")[0] if "-" in locale else locale
        rule = cls.RULES.get(lang, lambda n: "one" if n == 1 else "other")
        return rule(count)


# Translation service
class TranslationService:
    """Translation service."""
    
    def __init__(
        self,
        translation_store: Optional[TranslationStore] = None,
        locale_store: Optional[LocaleStore] = None,
        default_locale: str = "en",
        fallback_locale: str = "en",
    ):
        self._translations = translation_store or InMemoryTranslationStore()
        self._locales = locale_store or InMemoryLocaleStore()
        self._default_locale = default_locale
        self._fallback_locale = fallback_locale
        self._current_locale = default_locale
        self._stats = TranslationStats()
        self._cache: Dict[str, Dict[str, str]] = {}
    
    # Locale management
    @property
    def locale(self) -> str:
        """Get current locale."""
        return self._current_locale
    
    async def set_locale(self, locale: str) -> None:
        """Set current locale."""
        locale_obj = await self._locales.get(locale)
        if not locale_obj:
            # Try fallback to language code
            lang = locale.split("-")[0]
            locale_obj = await self._locales.get(lang)
        
        if locale_obj:
            self._current_locale = locale_obj.code
        else:
            logger.warning(f"Locale not found: {locale}")
            self._current_locale = self._fallback_locale
    
    async def get_locale(self, code: Optional[str] = None) -> Optional[Locale]:
        """Get locale."""
        return await self._locales.get(code or self._current_locale)
    
    async def list_locales(self) -> List[Locale]:
        """List available locales."""
        return await self._locales.list()
    
    async def add_locale(
        self,
        code: str,
        name: str,
        native_name: str = "",
        **kwargs,
    ) -> Locale:
        """Add locale."""
        locale = Locale(
            code=code,
            name=name,
            native_name=native_name or name,
            language=code.split("-")[0],
            **kwargs,
        )
        await self._locales.save(locale)
        self._stats.total_locales += 1
        return locale
    
    # Translation management
    async def add_translation(
        self,
        key: str,
        value: str,
        locale: Optional[str] = None,
        plurals: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Translation:
        """Add translation."""
        loc = locale or self._current_locale
        
        translation = Translation(
            key=key,
            value=value,
            locale=loc,
            plurals=plurals or {},
            **kwargs,
        )
        
        await self._translations.save_translation(loc, key, translation)
        
        # Clear cache
        self._cache.pop(loc, None)
        
        self._stats.total_translations += 1
        self._stats.by_locale[loc] = self._stats.by_locale.get(loc, 0) + 1
        
        return translation
    
    async def add_translations(
        self,
        locale: str,
        translations: Dict[str, str],
    ) -> int:
        """Add multiple translations."""
        count = 0
        for key, value in translations.items():
            # Check for plural forms (pipe-separated)
            if "|" in value:
                parts = value.split("|")
                plurals = {}
                
                if len(parts) == 2:
                    plurals["one"] = parts[0]
                    plurals["other"] = parts[1]
                elif len(parts) >= 3:
                    plurals["zero"] = parts[0]
                    plurals["one"] = parts[1]
                    plurals["other"] = parts[2] if len(parts) > 2 else parts[1]
                
                await self.add_translation(
                    key=key,
                    value=parts[0],
                    locale=locale,
                    plurals=plurals,
                )
            else:
                await self.add_translation(
                    key=key,
                    value=value,
                    locale=locale,
                )
            count += 1
        
        return count
    
    async def get_translation(
        self,
        key: str,
        locale: Optional[str] = None,
    ) -> Optional[Translation]:
        """Get translation."""
        loc = locale or self._current_locale
        return await self._translations.get_translation(loc, key)
    
    async def get_all_translations(
        self,
        locale: Optional[str] = None,
    ) -> Dict[str, Translation]:
        """Get all translations for locale."""
        loc = locale or self._current_locale
        return await self._translations.get_all(loc)
    
    # Translation
    async def translate(
        self,
        key: str,
        locale: Optional[str] = None,
        count: Optional[int] = None,
        **kwargs,
    ) -> str:
        """Translate key."""
        loc = locale or self._current_locale
        
        # Get translation
        translation = await self._translations.get_translation(loc, key)
        
        # Fallback
        if not translation and loc != self._fallback_locale:
            translation = await self._translations.get_translation(
                self._fallback_locale, key
            )
        
        if not translation:
            self._stats.missing_translations += 1
            return key  # Return key if no translation found
        
        # Get appropriate form
        value = translation.value
        
        if count is not None and translation.plurals:
            plural_form = PluralRules.get_plural_form(loc, count)
            value = translation.plurals.get(plural_form, translation.value)
        
        # Interpolate
        value = self._interpolate(value, count=count, **kwargs)
        
        return value
    
    async def t(
        self,
        key: str,
        **kwargs,
    ) -> str:
        """Shorthand for translate."""
        return await self.translate(key, **kwargs)
    
    def _interpolate(self, text: str, **kwargs) -> str:
        """Interpolate variables."""
        for key, value in kwargs.items():
            if value is not None:
                text = text.replace(f"{{{key}}}", str(value))
        return text
    
    # Formatting
    async def format_number(
        self,
        number: float,
        locale: Optional[str] = None,
    ) -> str:
        """Format number."""
        loc = await self.get_locale(locale)
        if not loc:
            return str(number)
        
        fmt = loc.number_format
        decimal = fmt.get("decimal", ".")
        thousands = fmt.get("thousands", ",")
        
        # Split into parts
        if isinstance(number, float):
            int_part, dec_part = str(number).split(".")
        else:
            int_part = str(number)
            dec_part = None
        
        # Add thousands separator
        int_part = "{:,}".format(int(int_part)).replace(",", thousands)
        
        if dec_part:
            return f"{int_part}{decimal}{dec_part}"
        return int_part
    
    async def format_currency(
        self,
        amount: float,
        currency: Optional[str] = None,
        locale: Optional[str] = None,
    ) -> str:
        """Format currency."""
        loc = await self.get_locale(locale)
        if not loc:
            return f"${amount:.2f}"
        
        curr = currency or loc.currency
        formatted = await self.format_number(amount, locale)
        
        # Apply currency format
        result = loc.currency_format.replace("{amount}", formatted)
        result = result.replace("$", curr)
        
        return result
    
    async def format_date(
        self,
        date: datetime,
        format: Optional[str] = None,
        locale: Optional[str] = None,
    ) -> str:
        """Format date."""
        loc = await self.get_locale(locale)
        if not loc:
            return date.strftime("%Y-%m-%d")
        
        fmt = format or loc.date_format
        return date.strftime(fmt)
    
    async def format_time(
        self,
        time: datetime,
        format: Optional[str] = None,
        locale: Optional[str] = None,
    ) -> str:
        """Format time."""
        loc = await self.get_locale(locale)
        if not loc:
            return time.strftime("%H:%M:%S")
        
        fmt = format or loc.time_format
        return time.strftime(fmt)
    
    async def format_datetime(
        self,
        dt: datetime,
        format: Optional[str] = None,
        locale: Optional[str] = None,
    ) -> str:
        """Format datetime."""
        loc = await self.get_locale(locale)
        if not loc:
            return dt.strftime("%Y-%m-%d %H:%M:%S")
        
        fmt = format or loc.datetime_format
        return dt.strftime(fmt)
    
    # Import/export
    async def import_json(
        self,
        locale: str,
        data: Dict[str, str],
    ) -> int:
        """Import translations from JSON."""
        return await self.add_translations(locale, data)
    
    async def export_json(
        self,
        locale: Optional[str] = None,
    ) -> Dict[str, str]:
        """Export translations to JSON."""
        translations = await self.get_all_translations(locale)
        return {
            key: t.value
            for key, t in translations.items()
        }
    
    # Stats
    def get_stats(self) -> TranslationStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_translation_service(
    default_locale: str = "en",
    fallback_locale: str = "en",
) -> TranslationService:
    """Create translation service."""
    return TranslationService(
        default_locale=default_locale,
        fallback_locale=fallback_locale,
    )


def create_locale(
    code: str,
    name: str,
    native_name: str = "",
    **kwargs,
) -> Locale:
    """Create locale."""
    return Locale(
        code=code,
        name=name,
        native_name=native_name or name,
        language=code.split("-")[0],
        **kwargs,
    )


def create_translation(
    key: str,
    value: str,
    locale: str = "en",
    **kwargs,
) -> Translation:
    """Create translation."""
    return Translation(
        key=key,
        value=value,
        locale=locale,
        **kwargs,
    )


__all__ = [
    # Exceptions
    "TranslationError",
    "LocaleNotFoundError",
    "TranslationNotFoundError",
    # Enums
    "PluralRule",
    # Data classes
    "Locale",
    "Translation",
    "TranslationBundle",
    "TranslationStats",
    # Stores
    "TranslationStore",
    "InMemoryTranslationStore",
    "LocaleStore",
    "InMemoryLocaleStore",
    # Rules
    "PluralRules",
    # Service
    "TranslationService",
    # Factory functions
    "create_translation_service",
    "create_locale",
    "create_translation",
]



"""
Enterprise API Documentation Module.

OpenAPI/Swagger generation, automatic API documentation,
endpoint discovery, and schema generation.

Example:
    # Create API documentation generator
    docs = create_api_docs(
        title="My API",
        version="1.0.0",
    )
    
    # Register endpoints
    docs.add_endpoint(
        path="/users",
        method="GET",
        summary="List users",
        responses={200: {"schema": {"type": "array"}}},
    )
    
    # Generate OpenAPI spec
    spec = await docs.generate_openapi()
    
    # Export as JSON/YAML
    json_spec = docs.to_json()
"""

from __future__ import annotations

import asyncio
import inspect
import json
import logging
import re
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class APIDocsError(Exception):
    """API documentation error."""
    pass


class HTTPMethod(str, Enum):
    """HTTP methods."""
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"
    HEAD = "head"
    OPTIONS = "options"


class ParameterLocation(str, Enum):
    """Parameter location."""
    PATH = "path"
    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"
    BODY = "body"


class DataType(str, Enum):
    """Data types."""
    STRING = "string"
    INTEGER = "integer"
    NUMBER = "number"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"


@dataclass
class Contact:
    """API contact info."""
    name: str = ""
    email: str = ""
    url: str = ""


@dataclass
class License:
    """API license info."""
    name: str = "MIT"
    url: str = ""


@dataclass
class ServerInfo:
    """Server info."""
    url: str = ""
    description: str = ""
    variables: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Schema:
    """JSON Schema."""
    type: DataType = DataType.OBJECT
    title: str = ""
    description: str = ""
    properties: Dict[str, Any] = field(default_factory=dict)
    required: List[str] = field(default_factory=list)
    items: Optional[Dict[str, Any]] = None
    enum: Optional[List[Any]] = None
    format: str = ""
    example: Any = None
    default: Any = None
    minimum: Optional[float] = None
    maximum: Optional[float] = None
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: str = ""
    nullable: bool = False
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result: Dict[str, Any] = {"type": self.type.value}
        
        if self.title:
            result["title"] = self.title
        if self.description:
            result["description"] = self.description
        if self.properties:
            result["properties"] = self.properties
        if self.required:
            result["required"] = self.required
        if self.items:
            result["items"] = self.items
        if self.enum:
            result["enum"] = self.enum
        if self.format:
            result["format"] = self.format
        if self.example is not None:
            result["example"] = self.example
        if self.default is not None:
            result["default"] = self.default
        if self.minimum is not None:
            result["minimum"] = self.minimum
        if self.maximum is not None:
            result["maximum"] = self.maximum
        if self.min_length is not None:
            result["minLength"] = self.min_length
        if self.max_length is not None:
            result["maxLength"] = self.max_length
        if self.pattern:
            result["pattern"] = self.pattern
        if self.nullable:
            result["nullable"] = self.nullable
        
        return result


@dataclass
class Parameter:
    """API parameter."""
    name: str = ""
    location: ParameterLocation = ParameterLocation.QUERY
    description: str = ""
    required: bool = False
    schema: Optional[Schema] = None
    example: Any = None
    deprecated: bool = False
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {
            "name": self.name,
            "in": self.location.value,
            "required": self.required,
        }
        
        if self.description:
            result["description"] = self.description
        
        if self.schema:
            result["schema"] = self.schema.to_dict()
        
        if self.example is not None:
            result["example"] = self.example
        
        if self.deprecated:
            result["deprecated"] = self.deprecated
        
        return result


@dataclass
class RequestBody:
    """Request body."""
    description: str = ""
    required: bool = True
    content: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {"required": self.required}
        
        if self.description:
            result["description"] = self.description
        
        if self.content:
            result["content"] = self.content
        
        return result


@dataclass
class Response:
    """API response."""
    status_code: int = 200
    description: str = ""
    schema: Optional[Schema] = None
    content_type: str = "application/json"
    headers: Dict[str, Any] = field(default_factory=dict)
    examples: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result: Dict[str, Any] = {}
        
        if self.description:
            result["description"] = self.description
        
        if self.schema:
            result["content"] = {
                self.content_type: {
                    "schema": self.schema.to_dict(),
                }
            }
            
            if self.examples:
                result["content"][self.content_type]["examples"] = self.examples
        
        if self.headers:
            result["headers"] = self.headers
        
        return result


@dataclass
class Endpoint:
    """API endpoint."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    path: str = ""
    method: HTTPMethod = HTTPMethod.GET
    summary: str = ""
    description: str = ""
    operation_id: str = ""
    tags: List[str] = field(default_factory=list)
    parameters: List[Parameter] = field(default_factory=list)
    request_body: Optional[RequestBody] = None
    responses: Dict[int, Response] = field(default_factory=dict)
    deprecated: bool = False
    security: List[Dict[str, List[str]]] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result: Dict[str, Any] = {}
        
        if self.summary:
            result["summary"] = self.summary
        if self.description:
            result["description"] = self.description
        if self.operation_id:
            result["operationId"] = self.operation_id
        if self.tags:
            result["tags"] = self.tags
        if self.deprecated:
            result["deprecated"] = self.deprecated
        
        if self.parameters:
            result["parameters"] = [p.to_dict() for p in self.parameters]
        
        if self.request_body:
            result["requestBody"] = self.request_body.to_dict()
        
        if self.responses:
            result["responses"] = {
                str(code): resp.to_dict()
                for code, resp in self.responses.items()
            }
        else:
            result["responses"] = {"200": {"description": "Successful response"}}
        
        if self.security:
            result["security"] = self.security
        
        return result


@dataclass
class Tag:
    """API tag."""
    name: str = ""
    description: str = ""
    external_docs: Optional[Dict[str, str]] = None
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {"name": self.name}
        
        if self.description:
            result["description"] = self.description
        
        if self.external_docs:
            result["externalDocs"] = self.external_docs
        
        return result


@dataclass
class SecurityScheme:
    """Security scheme."""
    name: str = ""
    type: str = "apiKey"  # apiKey, http, oauth2, openIdConnect
    description: str = ""
    location: str = "header"  # query, header, cookie
    scheme: str = ""  # bearer, basic
    bearer_format: str = ""
    flows: Optional[Dict[str, Any]] = None
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {"type": self.type}
        
        if self.description:
            result["description"] = self.description
        
        if self.type == "apiKey":
            result["name"] = self.name
            result["in"] = self.location
        
        elif self.type == "http":
            result["scheme"] = self.scheme
            if self.bearer_format:
                result["bearerFormat"] = self.bearer_format
        
        elif self.type == "oauth2" and self.flows:
            result["flows"] = self.flows
        
        return result


@dataclass
class APIInfo:
    """API info."""
    title: str = "API"
    description: str = ""
    version: str = "1.0.0"
    terms_of_service: str = ""
    contact: Optional[Contact] = None
    license: Optional[License] = None
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {
            "title": self.title,
            "version": self.version,
        }
        
        if self.description:
            result["description"] = self.description
        if self.terms_of_service:
            result["termsOfService"] = self.terms_of_service
        if self.contact:
            result["contact"] = {
                k: v for k, v in {
                    "name": self.contact.name,
                    "email": self.contact.email,
                    "url": self.contact.url,
                }.items() if v
            }
        if self.license:
            result["license"] = {
                "name": self.license.name,
                "url": self.license.url,
            } if self.license.url else {"name": self.license.name}
        
        return result


@dataclass
class OpenAPISpec:
    """OpenAPI specification."""
    openapi: str = "3.0.3"
    info: APIInfo = field(default_factory=APIInfo)
    servers: List[ServerInfo] = field(default_factory=list)
    paths: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    tags: List[Tag] = field(default_factory=list)
    components: Dict[str, Any] = field(default_factory=dict)
    security: List[Dict[str, List[str]]] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict."""
        result = {
            "openapi": self.openapi,
            "info": self.info.to_dict(),
        }
        
        if self.servers:
            result["servers"] = [
                {"url": s.url, "description": s.description}
                for s in self.servers
            ]
        
        if self.paths:
            result["paths"] = self.paths
        
        if self.tags:
            result["tags"] = [t.to_dict() for t in self.tags]
        
        if self.components:
            result["components"] = self.components
        
        if self.security:
            result["security"] = self.security
        
        return result


# Schema registry
class SchemaRegistry:
    """Schema registry."""
    
    def __init__(self):
        self._schemas: Dict[str, Schema] = {}
    
    def register(self, name: str, schema: Schema) -> None:
        """Register schema."""
        self._schemas[name] = schema
    
    def get(self, name: str) -> Optional[Schema]:
        """Get schema."""
        return self._schemas.get(name)
    
    def get_ref(self, name: str) -> Dict[str, str]:
        """Get schema reference."""
        return {"$ref": f"#/components/schemas/{name}"}
    
    def all(self) -> Dict[str, Schema]:
        """Get all schemas."""
        return self._schemas.copy()
    
    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        """Convert all to dict."""
        return {name: schema.to_dict() for name, schema in self._schemas.items()}


# Type to schema converter
class TypeConverter:
    """Python type to JSON schema converter."""
    
    _type_map = {
        str: DataType.STRING,
        int: DataType.INTEGER,
        float: DataType.NUMBER,
        bool: DataType.BOOLEAN,
        list: DataType.ARRAY,
        dict: DataType.OBJECT,
    }
    
    @classmethod
    def convert(cls, python_type: Type) -> Schema:
        """Convert Python type to schema."""
        origin = getattr(python_type, "__origin__", None)
        
        if origin is list:
            args = getattr(python_type, "__args__", (Any,))
            item_schema = cls.convert(args[0]) if args else Schema(type=DataType.STRING)
            return Schema(type=DataType.ARRAY, items=item_schema.to_dict())
        
        if origin is dict:
            return Schema(type=DataType.OBJECT)
        
        if python_type in cls._type_map:
            return Schema(type=cls._type_map[python_type])
        
        return Schema(type=DataType.STRING)


# API documentation generator
class APIDocumentation:
    """API documentation generator."""
    
    def __init__(
        self,
        title: str = "API",
        version: str = "1.0.0",
        description: str = "",
    ):
        self._info = APIInfo(
            title=title,
            version=version,
            description=description,
        )
        self._servers: List[ServerInfo] = []
        self._endpoints: List[Endpoint] = []
        self._tags: List[Tag] = []
        self._security_schemes: Dict[str, SecurityScheme] = {}
        self._schema_registry = SchemaRegistry()
        self._global_security: List[Dict[str, List[str]]] = []
    
    def set_info(
        self,
        title: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        contact: Optional[Contact] = None,
        license: Optional[License] = None,
    ) -> None:
        """Set API info."""
        if title:
            self._info.title = title
        if version:
            self._info.version = version
        if description:
            self._info.description = description
        if contact:
            self._info.contact = contact
        if license:
            self._info.license = license
    
    def add_server(
        self,
        url: str,
        description: str = "",
    ) -> None:
        """Add server."""
        self._servers.append(ServerInfo(url=url, description=description))
    
    def add_tag(
        self,
        name: str,
        description: str = "",
    ) -> None:
        """Add tag."""
        self._tags.append(Tag(name=name, description=description))
    
    def add_schema(
        self,
        name: str,
        schema: Schema,
    ) -> None:
        """Add reusable schema."""
        self._schema_registry.register(name, schema)
    
    def add_security_scheme(
        self,
        name: str,
        scheme: SecurityScheme,
    ) -> None:
        """Add security scheme."""
        self._security_schemes[name] = scheme
    
    def add_endpoint(
        self,
        path: str,
        method: Union[str, HTTPMethod],
        summary: str = "",
        description: str = "",
        tags: Optional[List[str]] = None,
        parameters: Optional[List[Dict[str, Any]]] = None,
        request_body: Optional[Dict[str, Any]] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        deprecated: bool = False,
        operation_id: str = "",
        security: Optional[List[Dict[str, List[str]]]] = None,
    ) -> Endpoint:
        """Add endpoint."""
        if isinstance(method, str):
            method = HTTPMethod(method.lower())
        
        # Convert parameters
        param_objs = []
        if parameters:
            for p in parameters:
                schema = None
                if "schema" in p:
                    schema = Schema(**p["schema"]) if isinstance(p["schema"], dict) else p["schema"]
                
                param_objs.append(Parameter(
                    name=p.get("name", ""),
                    location=ParameterLocation(p.get("in", "query")),
                    description=p.get("description", ""),
                    required=p.get("required", False),
                    schema=schema,
                    example=p.get("example"),
                ))
        
        # Convert request body
        req_body = None
        if request_body:
            req_body = RequestBody(
                description=request_body.get("description", ""),
                required=request_body.get("required", True),
                content=request_body.get("content", {}),
            )
        
        # Convert responses
        resp_objs = {}
        if responses:
            for code, resp in responses.items():
                schema = None
                if "schema" in resp:
                    schema_data = resp["schema"]
                    if isinstance(schema_data, dict):
                        schema = Schema(**{k: v for k, v in schema_data.items() if k != "type"})
                        schema.type = DataType(schema_data.get("type", "object"))
                
                resp_objs[code] = Response(
                    status_code=code,
                    description=resp.get("description", ""),
                    schema=schema,
                )
        
        endpoint = Endpoint(
            path=path,
            method=method,
            summary=summary,
            description=description,
            operation_id=operation_id or f"{method.value}_{path.replace('/', '_').strip('_')}",
            tags=tags or [],
            parameters=param_objs,
            request_body=req_body,
            responses=resp_objs,
            deprecated=deprecated,
            security=security or [],
        )
        
        self._endpoints.append(endpoint)
        
        logger.info(f"Endpoint added: {method.value.upper()} {path}")
        
        return endpoint
    
    def document_function(
        self,
        func: Callable,
        path: str,
        method: Union[str, HTTPMethod] = HTTPMethod.GET,
        tags: Optional[List[str]] = None,
    ) -> Endpoint:
        """Document function as endpoint."""
        summary = func.__name__.replace("_", " ").title()
        description = func.__doc__ or ""
        
        # Extract parameters from function signature
        sig = inspect.signature(func)
        parameters = []
        
        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls", "request"):
                continue
            
            schema_type = DataType.STRING
            if param.annotation != inspect.Parameter.empty:
                try:
                    schema = TypeConverter.convert(param.annotation)
                    schema_type = schema.type
                except Exception:
                    pass
            
            parameters.append({
                "name": param_name,
                "in": "query",
                "required": param.default == inspect.Parameter.empty,
                "schema": {"type": schema_type.value},
            })
        
        return self.add_endpoint(
            path=path,
            method=method,
            summary=summary,
            description=description.strip(),
            tags=tags,
            parameters=parameters,
        )
    
    async def generate_openapi(self) -> OpenAPISpec:
        """Generate OpenAPI specification."""
        # Build paths
        paths: Dict[str, Dict[str, Any]] = {}
        
        for endpoint in self._endpoints:
            if endpoint.path not in paths:
                paths[endpoint.path] = {}
            
            paths[endpoint.path][endpoint.method.value] = endpoint.to_dict()
        
        # Build components
        components: Dict[str, Any] = {}
        
        schemas = self._schema_registry.to_dict()
        if schemas:
            components["schemas"] = schemas
        
        if self._security_schemes:
            components["securitySchemes"] = {
                name: scheme.to_dict()
                for name, scheme in self._security_schemes.items()
            }
        
        spec = OpenAPISpec(
            info=self._info,
            servers=self._servers,
            paths=paths,
            tags=self._tags,
            components=components,
            security=self._global_security,
        )
        
        return spec
    
    def to_dict(self) -> Dict[str, Any]:
        """Synchronously generate spec dict."""
        loop = asyncio.new_event_loop()
        try:
            spec = loop.run_until_complete(self.generate_openapi())
            return spec.to_dict()
        finally:
            loop.close()
    
    def to_json(self, indent: int = 2) -> str:
        """Export as JSON."""
        return json.dumps(self.to_dict(), indent=indent)
    
    def to_yaml(self) -> str:
        """Export as YAML."""
        try:
            import yaml
            return yaml.dump(self.to_dict(), default_flow_style=False)
        except ImportError:
            raise APIDocsError("PyYAML required for YAML export")
    
    def get_endpoints(self) -> List[Endpoint]:
        """Get all endpoints."""
        return self._endpoints.copy()


# Decorator for documenting endpoints
def document(
    path: str,
    method: Union[str, HTTPMethod] = HTTPMethod.GET,
    summary: str = "",
    tags: Optional[List[str]] = None,
    docs: Optional[APIDocumentation] = None,
):
    """Decorator to document endpoint."""
    def decorator(func: Callable) -> Callable:
        if docs:
            docs.document_function(func, path, method, tags)
        
        func._api_path = path
        func._api_method = method
        func._api_summary = summary
        func._api_tags = tags
        
        return func
    
    return decorator


# Factory functions
def create_api_docs(
    title: str = "API",
    version: str = "1.0.0",
    description: str = "",
) -> APIDocumentation:
    """Create API documentation."""
    return APIDocumentation(title=title, version=version, description=description)


def create_schema(
    type: Union[str, DataType] = DataType.OBJECT,
    **kwargs,
) -> Schema:
    """Create schema."""
    if isinstance(type, str):
        type = DataType(type)
    return Schema(type=type, **kwargs)


def create_parameter(
    name: str,
    location: Union[str, ParameterLocation] = ParameterLocation.QUERY,
    **kwargs,
) -> Parameter:
    """Create parameter."""
    if isinstance(location, str):
        location = ParameterLocation(location)
    return Parameter(name=name, location=location, **kwargs)


def create_response(
    status_code: int = 200,
    description: str = "",
    schema: Optional[Schema] = None,
) -> Response:
    """Create response."""
    return Response(status_code=status_code, description=description, schema=schema)


__all__ = [
    # Exceptions
    "APIDocsError",
    # Enums
    "HTTPMethod",
    "ParameterLocation",
    "DataType",
    # Data classes
    "Contact",
    "License",
    "ServerInfo",
    "Schema",
    "Parameter",
    "RequestBody",
    "Response",
    "Endpoint",
    "Tag",
    "SecurityScheme",
    "APIInfo",
    "OpenAPISpec",
    # Classes
    "SchemaRegistry",
    "TypeConverter",
    "APIDocumentation",
    # Decorators
    "document",
    # Factory functions
    "create_api_docs",
    "create_schema",
    "create_parameter",
    "create_response",
]



"""
LLM Provider Base Interface.

Defines the common interface for all LLM provider adapters.
"""

import os
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

logger = logging.getLogger(__name__)


@dataclass
class LLMMessage:
    """Represents a message in a conversation."""
    role: str  # 'system', 'user', 'assistant', 'tool'
    content: str
    name: Optional[str] = None
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[Dict[str, Any]]] = None
    
    def to_dict(self) -> Dict[str, Any]:
        d = {'role': self.role, 'content': self.content}
        if self.name:
            d['name'] = self.name
        if self.tool_call_id:
            d['tool_call_id'] = self.tool_call_id
        if self.tool_calls:
            d['tool_calls'] = self.tool_calls
        return d


@dataclass
class LLMResponse:
    """Standardized response from LLM providers."""
    content: str
    model: str
    provider: str
    finish_reason: str = "stop"
    tool_calls: Optional[List[Dict[str, Any]]] = None
    usage: Dict[str, int] = field(default_factory=dict)
    raw_response: Optional[Any] = None
    
    @property
    def has_tool_calls(self) -> bool:
        return bool(self.tool_calls)


@dataclass
class ProviderConfig:
    """Configuration for an LLM provider."""
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    organization: Optional[str] = None
    default_model: Optional[str] = None
    timeout: float = 60.0
    max_retries: int = 3
    extra: Dict[str, Any] = field(default_factory=dict)
    
    @classmethod
    def from_env(cls, prefix: str = "") -> 'ProviderConfig':
        """Create config from environment variables."""
        return cls(
            api_key=os.getenv(f"{prefix}API_KEY"),
            api_base=os.getenv(f"{prefix}API_BASE"),
            organization=os.getenv(f"{prefix}ORGANIZATION"),
            default_model=os.getenv(f"{prefix}DEFAULT_MODEL"),
            timeout=float(os.getenv(f"{prefix}TIMEOUT", "60")),
            max_retries=int(os.getenv(f"{prefix}MAX_RETRIES", "3")),
        )


class BaseLLMProvider(ABC):
    """
    Base class for LLM provider adapters.
    
    All providers must implement:
    - generate(): Synchronous text generation
    - generate_with_tools(): Tool-augmented generation
    
    Optional:
    - stream(): Streaming generation
    - agenerate(): Async generation
    """
    
    def __init__(self, config: Optional[ProviderConfig] = None):
        self.config = config or ProviderConfig()
        self._client: Any = None
        self._initialized = False
    
    @property
    @abstractmethod
    def provider_name(self) -> str:
        """Name of the provider (e.g., 'openai', 'anthropic')."""
        pass
    
    @property
    @abstractmethod
    def supported_models(self) -> List[str]:
        """List of supported model identifiers."""
        pass
    
    @abstractmethod
    def _initialize_client(self) -> None:
        """Initialize the provider's client library."""
        pass
    
    def _ensure_initialized(self) -> None:
        """Ensure the client is initialized."""
        if not self._initialized:
            self._initialize_client()
            self._initialized = True
    
    @abstractmethod
    def generate(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        stop: Optional[List[str]] = None,
        **kwargs,
    ) -> LLMResponse:
        """
        Generate a response from the LLM.
        
        Args:
            prompt: The input prompt
            model: Model to use (defaults to config.default_model)
            temperature: Sampling temperature
            max_tokens: Maximum tokens to generate
            stop: Stop sequences
            **kwargs: Provider-specific parameters
            
        Returns:
            LLMResponse with the generated content
        """
        pass
    
    @abstractmethod
    def generate_with_tools(
        self,
        prompt: str,
        tools: List[Dict[str, Any]],
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        **kwargs,
    ) -> LLMResponse:
        """
        Generate a response with tool/function calling support.
        
        Args:
            prompt: The input prompt
            tools: List of tool schemas (OpenAI function format)
            model: Model to use
            temperature: Sampling temperature
            **kwargs: Provider-specific parameters
            
        Returns:
            LLMResponse with content and/or tool_calls
        """
        pass
    
    def generate_chat(
        self,
        messages: List[LLMMessage],
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        **kwargs,
    ) -> LLMResponse:
        """
        Generate from a list of messages (chat format).
        
        Default implementation converts to single prompt.
        Override for native chat support.
        """
        # Simple conversion - providers should override
        prompt = "\n".join(f"{m.role}: {m.content}" for m in messages)
        return self.generate(prompt, model=model, temperature=temperature, 
                           max_tokens=max_tokens, **kwargs)
    
    def stream(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        **kwargs,
    ) -> Iterator[str]:
        """
        Stream tokens as they are generated.
        
        Default: yields complete response at once.
        Override for true streaming.
        """
        response = self.generate(prompt, model=model, temperature=temperature, **kwargs)
        yield response.content
    
    async def agenerate(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        **kwargs,
    ) -> LLMResponse:
        """
        Async generation. Default runs sync in executor.
        Override for native async support.
        """
        import asyncio
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, 
            lambda: self.generate(prompt, model=model, temperature=temperature, **kwargs)
        )
    
    def count_tokens(self, text: str, model: Optional[str] = None) -> int:
        """
        Count tokens in text. Default uses rough estimate.
        Override for accurate counting.
        """
        # Rough estimate: 1 token ≈ 4 characters
        return len(text) // 4
    
    def is_available(self) -> bool:
        """Check if the provider is configured and available."""
        return bool(self.config.api_key)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(provider={self.provider_name})"


__all__ = [
    'LLMMessage',
    'LLMResponse', 
    'ProviderConfig',
    'BaseLLMProvider',
]



"""
AI Framework Integration Tools (LlamaIndex, LangChain).
"""

import logging
from typing import Any, Dict, List, Optional

from ..base import BaseTool, ToolConfig

logger = logging.getLogger(__name__)


class LlamaIndexTool(BaseTool):
    """
    Tool for integration with LlamaIndex.
    
    Features:
    - Index creation and querying
    - Document loading
    - Vector store integration
    - Custom retrievers
    """
    
    def __init__(
        self,
        config: Optional[ToolConfig] = None,
        api_key: Optional[str] = None,
    ):
        super().__init__(config or ToolConfig(
            name="LlamaIndexTool",
            description="LlamaIndex integration for RAG"
        ))
        self.api_key = api_key or self.config.api_key
        self._index = None
        self._service_context = None
    
    def _execute(
        self,
        query: str,
        documents: Optional[List[str]] = None,
        index_type: str = 'vector',
        top_k: int = 5,
        response_mode: str = 'compact',
    ) -> Dict[str, Any]:
        """
        Query documents using LlamaIndex.
        
        Args:
            query: Query string
            documents: List of document paths or texts
            index_type: 'vector', 'list', 'tree', or 'keyword'
            top_k: Number of results
            response_mode: 'compact', 'refine', 'tree_summarize'
            
        Returns:
            Dict with query results
        """
        try:
            from llama_index.core import (
                VectorStoreIndex,
                SimpleDirectoryReader,
                Document,
                Settings,
            )
            from llama_index.llms.openai import OpenAI
        except ImportError:
            raise ImportError("LlamaIndex requires: pip install llama-index")
        
        # Set up LLM
        if self.api_key:
            Settings.llm = OpenAI(api_key=self.api_key)
        
        # Load or create index
        if documents and (not self._index or documents):
            self._index = self._create_index(documents, index_type)
        
        if not self._index:
            return {
                'query': query,
                'status': 'error',
                'error': 'No index available. Provide documents to index.',
            }
        
        # Query
        query_engine = self._index.as_query_engine(
            similarity_top_k=top_k,
            response_mode=response_mode,
        )
        
        response = query_engine.query(query)
        
        return {
            'query': query,
            'status': 'success',
            'response': str(response),
            'source_nodes': [
                {
                    'text': node.node.text[:500],
                    'score': node.score,
                }
                for node in response.source_nodes
            ],
        }
    
    def _create_index(self, documents: List[str], index_type: str):
        """Create index from documents."""
        from llama_index.core import (
            VectorStoreIndex,
            ListIndex,
            TreeIndex,
            KeywordTableIndex,
            Document,
            SimpleDirectoryReader,
        )
        from pathlib import Path
        
        # Load documents
        docs = []
        for doc in documents:
            if Path(doc).exists():
                if Path(doc).is_dir():
                    reader = SimpleDirectoryReader(doc)
                    docs.extend(reader.load_data())
                else:
                    reader = SimpleDirectoryReader(input_files=[doc])
                    docs.extend(reader.load_data())
            else:
                # Treat as text
                docs.append(Document(text=doc))
        
        # Create index
        index_classes = {
            'vector': VectorStoreIndex,
            'list': ListIndex,
            'tree': TreeIndex,
            'keyword': KeywordTableIndex,
        }
        
        index_class = index_classes.get(index_type, VectorStoreIndex)
        return index_class.from_documents(docs)
    
    def add_documents(self, documents: List[str]):
        """Add documents to existing index."""
        from llama_index.core import Document, SimpleDirectoryReader
        from pathlib import Path
        
        if not self._index:
            self._index = self._create_index(documents, 'vector')
            return
        
        for doc in documents:
            if Path(doc).exists():
                reader = SimpleDirectoryReader(input_files=[doc])
                nodes = reader.load_data()
                for node in nodes:
                    self._index.insert(node)
            else:
                self._index.insert(Document(text=doc))


class LangChainTool(BaseTool):
    """
    Tool for integration with LangChain.
    
    Features:
    - Chain execution
    - Agent integration
    - Memory management
    - Tool wrapping
    """
    
    def __init__(
        self,
        config: Optional[ToolConfig] = None,
        api_key: Optional[str] = None,
    ):
        super().__init__(config or ToolConfig(
            name="LangChainTool",
            description="LangChain integration for AI chains"
        ))
        self.api_key = api_key or self.config.api_key
        self._chains: Dict[str, Any] = {}
    
    def _execute(
        self,
        prompt: str,
        chain_type: str = 'llm',
        memory: bool = False,
        tools: Optional[List[str]] = None,
        model: str = 'gpt-4',
    ) -> Dict[str, Any]:
        """
        Execute LangChain operation.
        
        Args:
            prompt: Input prompt
            chain_type: 'llm', 'conversation', 'agent', 'qa'
            memory: Enable conversation memory
            tools: List of tool names for agent
            model: Model to use
            
        Returns:
            Dict with execution results
        """
        try:
            from langchain_openai import ChatOpenAI
            from langchain.chains import LLMChain, ConversationChain
            from langchain.memory import ConversationBufferMemory
            from langchain.prompts import PromptTemplate
        except ImportError:
            raise ImportError("LangChain requires: pip install langchain langchain-openai")
        
        if not self.api_key:
            raise ValueError("API key required")
        
        llm = ChatOpenAI(model=model, api_key=self.api_key)
        
        if chain_type == 'llm':
            result = self._run_llm_chain(llm, prompt)
        elif chain_type == 'conversation':
            result = self._run_conversation(llm, prompt, memory)
        elif chain_type == 'agent':
            result = self._run_agent(llm, prompt, tools)
        else:
            result = {'error': f'Unknown chain type: {chain_type}'}
        
        return {
            'prompt': prompt,
            'chain_type': chain_type,
            'model': model,
            **result,
        }
    
    def _run_llm_chain(self, llm, prompt: str) -> Dict[str, Any]:
        """Run simple LLM chain."""
        from langchain.chains import LLMChain
        from langchain.prompts import PromptTemplate
        
        template = PromptTemplate(
            input_variables=["input"],
            template="{input}"
        )
        
        chain = LLMChain(llm=llm, prompt=template)
        result = chain.invoke({"input": prompt})
        
        return {
            'status': 'success',
            'response': result.get('text', str(result)),
        }
    
    def _run_conversation(self, llm, prompt: str, use_memory: bool) -> Dict[str, Any]:
        """Run conversation chain."""
        from langchain.chains import ConversationChain
        from langchain.memory import ConversationBufferMemory
        
        chain_id = 'conversation'
        
        if chain_id not in self._chains or not use_memory:
            memory = ConversationBufferMemory() if use_memory else None
            self._chains[chain_id] = ConversationChain(
                llm=llm,
                memory=memory,
                verbose=False,
            )
        
        chain = self._chains[chain_id]
        result = chain.invoke({"input": prompt})
        
        return {
            'status': 'success',
            'response': result.get('response', str(result)),
        }
    
    def _run_agent(self, llm, prompt: str, tool_names: Optional[List[str]]) -> Dict[str, Any]:
        """Run agent with tools."""
        try:
            from langchain.agents import AgentExecutor, create_openai_functions_agent
            from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
        except ImportError:
            return {'status': 'error', 'error': 'Agent support requires additional packages'}
        
        # For now, return a simple response
        # Full agent implementation would require tool definitions
        from langchain.chains import LLMChain
        from langchain.prompts import PromptTemplate
        
        template = PromptTemplate(
            input_variables=["input"],
            template="You are a helpful assistant. {input}"
        )
        
        chain = LLMChain(llm=llm, prompt=template)
        result = chain.invoke({"input": prompt})
        
        return {
            'status': 'success',
            'response': result.get('text', str(result)),
            'tools_available': tool_names or [],
        }
    
    def clear_memory(self):
        """Clear conversation memory."""
        self._chains.clear()


__all__ = ['LlamaIndexTool', 'LangChainTool']



"""
Output Formatter Implementation.

Provides comprehensive formatting for agent outputs including
Markdown, Code, JSON, HTML, Tables, and Plain Text.
"""

import json
import re
import html
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union, Callable

logger = logging.getLogger(__name__)


class FormatType(Enum):
    """Supported output format types."""
    MARKDOWN = "markdown"
    CODE = "code"
    JSON = "json"
    HTML = "html"
    TABLE = "table"
    PLAIN = "plain"
    XML = "xml"
    YAML = "yaml"


class TableFormat(Enum):
    """Table output formats."""
    MARKDOWN = "markdown"
    ASCII = "ascii"
    CSV = "csv"
    TSV = "tsv"
    HTML = "html"
    LATEX = "latex"


@dataclass
class OutputFormat:
    """Output format configuration."""
    format_type: FormatType
    options: Dict[str, Any] = field(default_factory=dict)
    template: Optional[str] = None
    syntax_highlight: bool = True
    line_numbers: bool = False
    max_width: int = 120
    indent: int = 2


@dataclass
class FormattedOutput:
    """Formatted output result."""
    content: str
    formatted: str
    format_type: FormatType
    metadata: Dict[str, Any] = field(default_factory=dict)
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    
    def __str__(self) -> str:
        return self.formatted
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "content": self.content,
            "formatted": self.formatted,
            "format_type": self.format_type.value,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }


@dataclass
class CodeBlock:
    """Code block with language and metadata."""
    code: str
    language: str = "text"
    filename: Optional[str] = None
    line_start: int = 1
    highlight_lines: List[int] = field(default_factory=list)
    
    def to_markdown(self) -> str:
        """Convert to markdown code block."""
        header = f"```{self.language}"
        if self.filename:
            header += f" title=\"{self.filename}\""
        return f"{header}\n{self.code}\n```"
    
    def to_html(self) -> str:
        """Convert to HTML code block."""
        escaped = html.escape(self.code)
        lang_class = f"language-{self.language}" if self.language else ""
        return f'<pre><code class="{lang_class}">{escaped}</code></pre>'


class BaseFormatter(ABC):
    """Abstract base formatter."""
    
    @abstractmethod
    def format(self, content: Any, **options) -> str:
        """Format content."""
        pass
    
    @abstractmethod
    def can_format(self, content: Any) -> bool:
        """Check if content can be formatted."""
        pass


class MarkdownFormatter(BaseFormatter):
    """
    Markdown formatter for rich text output.
    
    Supports:
    - Headings (h1-h6)
    - Lists (ordered, unordered, checklists)
    - Tables
    - Code blocks
    - Links and images
    - Blockquotes
    - Horizontal rules
    - Bold, italic, strikethrough
    """
    
    def can_format(self, content: Any) -> bool:
        return isinstance(content, (str, dict, list))
    
    def format(self, content: Any, **options) -> str:
        """Format content as markdown."""
        template = options.get("template", "paragraph")
        
        if template == "heading":
            level = options.get("level", 1)
            return self.heading(str(content), level)
        elif template == "list":
            return self.list_items(content if isinstance(content, list) else [content], 
                                   ordered=options.get("ordered", False))
        elif template == "checklist":
            return self.checklist(content if isinstance(content, list) else [content])
        elif template == "table":
            return self.table(content)
        elif template == "blockquote":
            return self.blockquote(str(content))
        elif template == "code":
            return self.code_block(str(content), options.get("language", ""))
        elif template == "link":
            return self.link(str(content), options.get("url", ""))
        else:
            return str(content)
    
    def heading(self, text: str, level: int = 1) -> str:
        """Create markdown heading."""
        level = max(1, min(6, level))
        return f"{'#' * level} {text}"
    
    def list_items(self, items: List[Any], ordered: bool = False, indent: int = 0) -> str:
        """Create markdown list."""
        lines = []
        prefix = "  " * indent
        for i, item in enumerate(items, 1):
            if isinstance(item, list):
                # Nested list
                lines.append(self.list_items(item, ordered, indent + 1))
            else:
                marker = f"{i}." if ordered else "-"
                lines.append(f"{prefix}{marker} {item}")
        return "\n".join(lines)
    
    def checklist(self, items: List[Union[str, tuple]]) -> str:
        """Create markdown checklist."""
        lines = []
        for item in items:
            if isinstance(item, tuple):
                text, checked = item
            else:
                text, checked = item, False
            checkbox = "[x]" if checked else "[ ]"
            lines.append(f"- {checkbox} {text}")
        return "\n".join(lines)
    
    def table(self, data: Union[List[Dict], Dict[str, List]]) -> str:
        """Create markdown table."""
        if not data:
            return ""
        
        if isinstance(data, list) and len(data) > 0 and isinstance(data[0], dict):
            # List of dicts
            headers = list(data[0].keys())
            rows = [[str(row.get(h, "")) for h in headers] for row in data]
        elif isinstance(data, dict):
            # Dict of lists
            headers = list(data.keys())
            max_len = max(len(v) if isinstance(v, list) else 1 for v in data.values())
            rows = []
            for i in range(max_len):
                row = []
                for h in headers:
                    vals = data[h] if isinstance(data[h], list) else [data[h]]
                    row.append(str(vals[i]) if i < len(vals) else "")
                rows.append(row)
        else:
            return str(data)
        
        # Build table
        header_line = "| " + " | ".join(headers) + " |"
        separator = "| " + " | ".join(["---"] * len(headers)) + " |"
        data_lines = ["| " + " | ".join(row) + " |" for row in rows]
        
        return "\n".join([header_line, separator] + data_lines)
    
    def blockquote(self, text: str) -> str:
        """Create markdown blockquote."""
        lines = text.split("\n")
        return "\n".join(f"> {line}" for line in lines)
    
    def code_block(self, code: str, language: str = "") -> str:
        """Create markdown code block."""
        return f"```{language}\n{code}\n```"
    
    def inline_code(self, code: str) -> str:
        """Create inline code."""
        return f"`{code}`"
    
    def link(self, text: str, url: str, title: str = "") -> str:
        """Create markdown link."""
        if title:
            return f'[{text}]({url} "{title}")'
        return f"[{text}]({url})"
    
    def image(self, alt: str, url: str, title: str = "") -> str:
        """Create markdown image."""
        if title:
            return f'![{alt}]({url} "{title}")'
        return f"![{alt}]({url})"
    
    def bold(self, text: str) -> str:
        return f"**{text}**"
    
    def italic(self, text: str) -> str:
        return f"*{text}*"
    
    def strikethrough(self, text: str) -> str:
        return f"~~{text}~~"
    
    def horizontal_rule(self) -> str:
        return "---"


class CodeFormatter(BaseFormatter):
    """
    Code formatter with syntax highlighting support.
    
    Features:
    - Language detection
    - Line numbers
    - Syntax highlighting (for terminal output)
    - Diff formatting
    """
    
    # Common language patterns for detection
    LANGUAGE_PATTERNS = {
        "python": [r"^\s*def\s+\w+", r"^\s*class\s+\w+", r"import\s+\w+", r"from\s+\w+\s+import"],
        "javascript": [r"^\s*function\s+\w+", r"^\s*const\s+\w+\s*=", r"^\s*let\s+\w+\s*=", r"=>\s*{"],
        "typescript": [r":\s*(string|number|boolean|any)\s*[;=]", r"interface\s+\w+", r"type\s+\w+\s*="],
        "java": [r"^\s*public\s+class", r"^\s*private\s+\w+", r"System\.out\.println"],
        "csharp": [r"^\s*namespace\s+", r"^\s*public\s+class", r"Console\.WriteLine"],
        "go": [r"^\s*func\s+\w+", r"^\s*package\s+\w+", r"fmt\.Println"],
        "rust": [r"^\s*fn\s+\w+", r"^\s*let\s+mut\s+", r"println!\s*\("],
        "sql": [r"^\s*SELECT\s+", r"^\s*INSERT\s+INTO", r"^\s*CREATE\s+TABLE"],
        "html": [r"<!DOCTYPE\s+html>", r"<html", r"<body"],
        "css": [r"^\s*\.\w+\s*{", r"^\s*#\w+\s*{", r"@media\s+"],
        "json": [r"^\s*\{", r'"\w+":\s*'],
        "yaml": [r"^\s*\w+:\s*\n", r"^\s*-\s+\w+"],
        "bash": [r"^#!/bin/bash", r"^\s*echo\s+", r"\$\(\w+\)"],
        "powershell": [r"^\$\w+\s*=", r"Get-\w+", r"Set-\w+"],
    }
    
    def can_format(self, content: Any) -> bool:
        return isinstance(content, str)
    
    def format(self, content: Any, **options) -> str:
        """Format code with options."""
        code = str(content)
        language = options.get("language") or self.detect_language(code)
        line_numbers = options.get("line_numbers", False)
        highlight_lines = options.get("highlight_lines", [])
        
        if line_numbers:
            code = self.add_line_numbers(code, highlight_lines)
        
        if options.get("markdown_wrap", True):
            return f"```{language}\n{code}\n```"
        
        return code
    
    def detect_language(self, code: str) -> str:
        """Detect programming language from code content."""
        for language, patterns in self.LANGUAGE_PATTERNS.items():
            for pattern in patterns:
                if re.search(pattern, code, re.MULTILINE | re.IGNORECASE):
                    return language
        return "text"
    
    def add_line_numbers(self, code: str, highlight_lines: List[int] = None) -> str:
        """Add line numbers to code."""
        highlight_lines = highlight_lines or []
        lines = code.split("\n")
        max_width = len(str(len(lines)))
        
        numbered_lines = []
        for i, line in enumerate(lines, 1):
            marker = ">" if i in highlight_lines else " "
            numbered_lines.append(f"{marker}{i:>{max_width}} | {line}")
        
        return "\n".join(numbered_lines)
    
    def format_diff(self, old_code: str, new_code: str, context_lines: int = 3) -> str:
        """Format code diff."""
        import difflib
        
        old_lines = old_code.splitlines(keepends=True)
        new_lines = new_code.splitlines(keepends=True)
        
        diff = difflib.unified_diff(
            old_lines, new_lines,
            fromfile="original",
            tofile="modified",
            n=context_lines
        )
        
        return "".join(diff)
    
    def extract_code_blocks(self, text: str) -> List[CodeBlock]:
        """Extract code blocks from markdown text."""
        pattern = r"```(\w*)\n(.*?)```"
        matches = re.findall(pattern, text, re.DOTALL)
        
        blocks = []
        for language, code in matches:
            blocks.append(CodeBlock(
                code=code.strip(),
                language=language or "text"
            ))
        
        return blocks


class JSONFormatter(BaseFormatter):
    """
    JSON formatter with pretty printing and validation.
    """
    
    def can_format(self, content: Any) -> bool:
        if isinstance(content, (dict, list)):
            return True
        if isinstance(content, str):
            try:
                json.loads(content)
                return True
            except json.JSONDecodeError:
                return False
        return False
    
    def format(self, content: Any, **options) -> str:
        """Format as pretty JSON."""
        indent = options.get("indent", 2)
        sort_keys = options.get("sort_keys", False)
        ensure_ascii = options.get("ensure_ascii", False)
        
        if isinstance(content, str):
            try:
                content = json.loads(content)
            except json.JSONDecodeError:
                return content
        
        return json.dumps(
            content,
            indent=indent,
            sort_keys=sort_keys,
            ensure_ascii=ensure_ascii,
            default=str
        )
    
    def minify(self, content: Any) -> str:
        """Minify JSON."""
        if isinstance(content, str):
            content = json.loads(content)
        return json.dumps(content, separators=(",", ":"))
    
    def validate(self, content: str, schema: Optional[Dict] = None) -> tuple:
        """Validate JSON and optionally against schema."""
        try:
            data = json.loads(content)
            if schema:
                try:
                    import jsonschema
                    jsonschema.validate(data, schema)
                except ImportError:
                    logger.warning("jsonschema not installed, skipping schema validation")
                except jsonschema.ValidationError as e:
                    return False, str(e)
            return True, None
        except json.JSONDecodeError as e:
            return False, str(e)


class HTMLFormatter(BaseFormatter):
    """
    HTML formatter for rich web output.
    """
    
    def can_format(self, content: Any) -> bool:
        return True
    
    def format(self, content: Any, **options) -> str:
        """Format content as HTML."""
        template = options.get("template", "paragraph")
        
        if template == "paragraph":
            return self.paragraph(str(content))
        elif template == "heading":
            level = options.get("level", 1)
            return self.heading(str(content), level)
        elif template == "list":
            return self.list_items(content if isinstance(content, list) else [content],
                                   ordered=options.get("ordered", False))
        elif template == "table":
            return self.table(content)
        elif template == "code":
            return self.code(str(content), options.get("language", ""))
        elif template == "card":
            return self.card(str(content), options.get("title", ""))
        else:
            return self.escape(str(content))
    
    def escape(self, text: str) -> str:
        """Escape HTML special characters."""
        return html.escape(text)
    
    def paragraph(self, text: str) -> str:
        return f"<p>{self.escape(text)}</p>"
    
    def heading(self, text: str, level: int = 1) -> str:
        level = max(1, min(6, level))
        return f"<h{level}>{self.escape(text)}</h{level}>"
    
    def list_items(self, items: List[Any], ordered: bool = False) -> str:
        tag = "ol" if ordered else "ul"
        items_html = "\n".join(f"  <li>{self.escape(str(item))}</li>" for item in items)
        return f"<{tag}>\n{items_html}\n</{tag}>"
    
    def table(self, data: Union[List[Dict], Dict]) -> str:
        """Create HTML table."""
        if not data:
            return "<table></table>"
        
        if isinstance(data, list) and len(data) > 0 and isinstance(data[0], dict):
            headers = list(data[0].keys())
            rows = [[str(row.get(h, "")) for h in headers] for row in data]
        else:
            return f"<table><tr><td>{self.escape(str(data))}</td></tr></table>"
        
        header_html = "<tr>" + "".join(f"<th>{self.escape(h)}</th>" for h in headers) + "</tr>"
        rows_html = "\n".join(
            "<tr>" + "".join(f"<td>{self.escape(cell)}</td>" for cell in row) + "</tr>"
            for row in rows
        )
        
        return f"<table>\n  <thead>\n    {header_html}\n  </thead>\n  <tbody>\n    {rows_html}\n  </tbody>\n</table>"
    
    def code(self, code: str, language: str = "") -> str:
        class_attr = f' class="language-{language}"' if language else ""
        return f"<pre><code{class_attr}>{self.escape(code)}</code></pre>"
    
    def card(self, content: str, title: str = "") -> str:
        title_html = f"<div class='card-title'>{self.escape(title)}</div>" if title else ""
        return f"""<div class="card">
  {title_html}
  <div class="card-content">{self.escape(content)}</div>
</div>"""
    
    def link(self, text: str, url: str, target: str = "_blank") -> str:
        return f'<a href="{self.escape(url)}" target="{target}">{self.escape(text)}</a>'
    
    def image(self, src: str, alt: str = "", width: str = "", height: str = "") -> str:
        attrs = [f'src="{self.escape(src)}"', f'alt="{self.escape(alt)}"']
        if width:
            attrs.append(f'width="{width}"')
        if height:
            attrs.append(f'height="{height}"')
        return f"<img {' '.join(attrs)} />"


class TableFormatter(BaseFormatter):
    """
    Table formatter supporting multiple output formats.
    """
    
    def can_format(self, content: Any) -> bool:
        return isinstance(content, (list, dict))
    
    def format(self, content: Any, **options) -> str:
        """Format data as table."""
        table_format = options.get("format", TableFormat.MARKDOWN)
        if isinstance(table_format, str):
            table_format = TableFormat(table_format)
        
        formatters = {
            TableFormat.MARKDOWN: self._format_markdown,
            TableFormat.ASCII: self._format_ascii,
            TableFormat.CSV: self._format_csv,
            TableFormat.TSV: self._format_tsv,
            TableFormat.HTML: self._format_html,
            TableFormat.LATEX: self._format_latex,
        }
        
        formatter = formatters.get(table_format, self._format_markdown)
        return formatter(content, **options)
    
    def _extract_data(self, content: Any) -> tuple:
        """Extract headers and rows from content."""
        if isinstance(content, list) and len(content) > 0:
            if isinstance(content[0], dict):
                headers = list(content[0].keys())
                rows = [[str(row.get(h, "")) for h in headers] for row in content]
            elif isinstance(content[0], (list, tuple)):
                headers = [f"Col{i+1}" for i in range(len(content[0]))]
                rows = [[str(cell) for cell in row] for row in content]
            else:
                headers = ["Value"]
                rows = [[str(item)] for item in content]
        elif isinstance(content, dict):
            headers = ["Key", "Value"]
            rows = [[str(k), str(v)] for k, v in content.items()]
        else:
            headers = ["Value"]
            rows = [[str(content)]]
        
        return headers, rows
    
    def _format_markdown(self, content: Any, **options) -> str:
        headers, rows = self._extract_data(content)
        
        header_line = "| " + " | ".join(headers) + " |"
        separator = "| " + " | ".join(["---"] * len(headers)) + " |"
        data_lines = ["| " + " | ".join(row) + " |" for row in rows]
        
        return "\n".join([header_line, separator] + data_lines)
    
    def _format_ascii(self, content: Any, **options) -> str:
        headers, rows = self._extract_data(content)
        
        # Calculate column widths
        widths = [len(h) for h in headers]
        for row in rows:
            for i, cell in enumerate(row):
                widths[i] = max(widths[i], len(cell))
        
        # Build table
        separator = "+" + "+".join("-" * (w + 2) for w in widths) + "+"
        
        def format_row(row):
            cells = [f" {cell:<{widths[i]}} " for i, cell in enumerate(row)]
            return "|" + "|".join(cells) + "|"
        
        lines = [separator, format_row(headers), separator]
        for row in rows:
            lines.append(format_row(row))
        lines.append(separator)
        
        return "\n".join(lines)
    
    def _format_csv(self, content: Any, **options) -> str:
        import csv
        import io
        
        headers, rows = self._extract_data(content)
        
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(headers)
        writer.writerows(rows)
        
        return output.getvalue()
    
    def _format_tsv(self, content: Any, **options) -> str:
        headers, rows = self._extract_data(content)
        
        lines = ["\t".join(headers)]
        for row in rows:
            lines.append("\t".join(row))
        
        return "\n".join(lines)
    
    def _format_html(self, content: Any, **options) -> str:
        html_formatter = HTMLFormatter()
        return html_formatter.table(content)
    
    def _format_latex(self, content: Any, **options) -> str:
        headers, rows = self._extract_data(content)
        
        col_format = "|" + "|".join(["c"] * len(headers)) + "|"
        
        lines = [
            "\\begin{tabular}{" + col_format + "}",
            "\\hline",
            " & ".join(headers) + " \\\\",
            "\\hline",
        ]
        
        for row in rows:
            lines.append(" & ".join(row) + " \\\\")
        
        lines.extend(["\\hline", "\\end{tabular}"])
        
        return "\n".join(lines)


class PlainTextFormatter(BaseFormatter):
    """
    Plain text formatter with word wrapping.
    """
    
    def can_format(self, content: Any) -> bool:
        return True
    
    def format(self, content: Any, **options) -> str:
        """Format as plain text."""
        text = str(content)
        max_width = options.get("max_width", 80)
        
        if options.get("wrap", False):
            text = self.wrap_text(text, max_width)
        
        return text
    
    def wrap_text(self, text: str, width: int = 80) -> str:
        """Wrap text at specified width."""
        import textwrap
        
        paragraphs = text.split("\n\n")
        wrapped = []
        
        for para in paragraphs:
            wrapped.append(textwrap.fill(para, width=width))
        
        return "\n\n".join(wrapped)
    
    def strip_formatting(self, text: str) -> str:
        """Remove markdown/HTML formatting."""
        # Remove markdown links
        text = re.sub(r'\[([^\]]+)\]\([^\)]+\)', r'\1', text)
        # Remove markdown bold/italic
        text = re.sub(r'\*\*([^\*]+)\*\*', r'\1', text)
        text = re.sub(r'\*([^\*]+)\*', r'\1', text)
        # Remove markdown code
        text = re.sub(r'`([^`]+)`', r'\1', text)
        # Remove HTML tags
        text = re.sub(r'<[^>]+>', '', text)
        
        return text


class OutputFormatter:
    """
    Main output formatter combining all format types.
    
    Example:
        >>> formatter = OutputFormatter()
        >>> 
        >>> # Format as markdown heading
        >>> result = formatter.format(
        ...     "Hello World",
        ...     format_type=FormatType.MARKDOWN,
        ...     template="heading",
        ...     level=1
        ... )
        >>> print(result.formatted)
        # Hello World
        >>> 
        >>> # Format code
        >>> code_result = formatter.format_code(
        ...     "print('hello')",
        ...     language="python",
        ...     line_numbers=True
        ... )
        >>> 
        >>> # Format table
        >>> table_result = formatter.format_table(
        ...     [{"name": "Alice", "age": 30}],
        ...     format=TableFormat.MARKDOWN
        ... )
    """
    
    def __init__(self):
        self._formatters: Dict[FormatType, BaseFormatter] = {
            FormatType.MARKDOWN: MarkdownFormatter(),
            FormatType.CODE: CodeFormatter(),
            FormatType.JSON: JSONFormatter(),
            FormatType.HTML: HTMLFormatter(),
            FormatType.TABLE: TableFormatter(),
            FormatType.PLAIN: PlainTextFormatter(),
        }
        
        # Custom formatters
        self._custom_formatters: Dict[str, Callable] = {}
    
    def register_formatter(self, name: str, formatter: Union[BaseFormatter, Callable]) -> None:
        """Register custom formatter."""
        if isinstance(formatter, BaseFormatter):
            self._custom_formatters[name] = formatter.format
        else:
            self._custom_formatters[name] = formatter
    
    def format(
        self,
        content: Any,
        format_type: FormatType = FormatType.PLAIN,
        **options
    ) -> FormattedOutput:
        """
        Format content with specified format type.
        
        Args:
            content: Content to format
            format_type: Output format type
            **options: Format-specific options
        
        Returns:
            FormattedOutput with original and formatted content
        """
        formatter = self._formatters.get(format_type, self._formatters[FormatType.PLAIN])
        
        try:
            formatted = formatter.format(content, **options)
            
            return FormattedOutput(
                content=str(content),
                formatted=formatted,
                format_type=format_type,
                metadata={"options": options}
            )
        except Exception as e:
            logger.error(f"Formatting error: {e}")
            return FormattedOutput(
                content=str(content),
                formatted=str(content),
                format_type=format_type,
                metadata={"error": str(e)}
            )
    
    def format_markdown(self, content: Any, **options) -> FormattedOutput:
        """Format as markdown."""
        return self.format(content, FormatType.MARKDOWN, **options)
    
    def format_code(
        self,
        code: str,
        language: str = "",
        line_numbers: bool = False,
        highlight_lines: List[int] = None,
        **options
    ) -> FormattedOutput:
        """Format code with syntax highlighting."""
        options.update({
            "language": language,
            "line_numbers": line_numbers,
            "highlight_lines": highlight_lines or [],
        })
        return self.format(code, FormatType.CODE, **options)
    
    def format_json(self, data: Any, indent: int = 2, sort_keys: bool = False) -> FormattedOutput:
        """Format as pretty JSON."""
        return self.format(data, FormatType.JSON, indent=indent, sort_keys=sort_keys)
    
    def format_html(self, content: Any, **options) -> FormattedOutput:
        """Format as HTML."""
        return self.format(content, FormatType.HTML, **options)
    
    def format_table(
        self,
        data: Union[List[Dict], Dict],
        format: TableFormat = TableFormat.MARKDOWN,
        **options
    ) -> FormattedOutput:
        """Format data as table."""
        options["format"] = format
        return self.format(data, FormatType.TABLE, **options)
    
    def format_plain(self, content: Any, max_width: int = 80, wrap: bool = False) -> FormattedOutput:
        """Format as plain text."""
        return self.format(content, FormatType.PLAIN, max_width=max_width, wrap=wrap)
    
    # Convenience methods for markdown
    def heading(self, text: str, level: int = 1) -> str:
        """Create markdown heading."""
        return self._formatters[FormatType.MARKDOWN].heading(text, level)
    
    def bullet_list(self, items: List[Any]) -> str:
        """Create markdown bullet list."""
        return self._formatters[FormatType.MARKDOWN].list_items(items, ordered=False)
    
    def numbered_list(self, items: List[Any]) -> str:
        """Create markdown numbered list."""
        return self._formatters[FormatType.MARKDOWN].list_items(items, ordered=True)
    
    def code_block(self, code: str, language: str = "") -> str:
        """Create markdown code block."""
        return self._formatters[FormatType.MARKDOWN].code_block(code, language)
    
    def table(self, data: Union[List[Dict], Dict]) -> str:
        """Create markdown table."""
        return self._formatters[FormatType.MARKDOWN].table(data)
    
    def blockquote(self, text: str) -> str:
        """Create markdown blockquote."""
        return self._formatters[FormatType.MARKDOWN].blockquote(text)


__all__ = [
    # Types
    "FormatType",
    "OutputFormat",
    "FormattedOutput",
    "CodeBlock",
    "TableFormat",
    # Formatters
    "BaseFormatter",
    "MarkdownFormatter",
    "CodeFormatter",
    "JSONFormatter",
    "HTMLFormatter",
    "TableFormatter",
    "PlainTextFormatter",
    # Main class
    "OutputFormatter",
]



"""
Task and Tool Evaluation systems.

Provides:
- TaskEvaluator: Task completion, instruction following, multi-step reasoning
- ToolInvocationEvaluator: Tool usage correctness, parameter validity, call ordering
"""

import uuid
import time
import logging
import statistics
from typing import Dict, Any, List
from collections import defaultdict

logger = logging.getLogger(__name__)


class TaskEvaluator:
    """Task/Skill-Level Evaluation system."""
    
    def __init__(self):
        self.task_executions: List[Dict[str, Any]] = []
        self.task_metrics: Dict[str, Dict[str, Any]] = defaultdict(
            lambda: {'attempts': 0, 'successes': 0, 'failures': 0, 'retries': []}
        )
    
    def record_task_execution(self,
                            task_name: str,
                            success: bool,
                            completion_percentage: float = None,
                            retry_count: int = 0,
                            error_recovered: bool = False,
                            duration_ms: float = None,
                            metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        """Record task execution."""
        execution = {
            'id': str(uuid.uuid4()),
            'task_name': task_name,
            'success': success,
            'completion_percentage': completion_percentage or (100.0 if success else 0.0),
            'retry_count': retry_count,
            'error_recovered': error_recovered,
            'duration_ms': duration_ms,
            'metadata': metadata or {},
            'timestamp': time.time()
        }
        
        self.task_executions.append(execution)
        
        metrics = self.task_metrics[task_name]
        metrics['attempts'] += 1
        if success:
            metrics['successes'] += 1
        else:
            metrics['failures'] += 1
        metrics['retries'].append(retry_count)
        
        return execution
    
    def get_task_metrics(self, task_name: str = None) -> Dict[str, Any]:
        """Get metrics for a specific task or all tasks."""
        if task_name:
            if task_name not in self.task_metrics:
                return {'error': 'Task not found'}
            
            metrics = self.task_metrics[task_name]
            total = metrics['attempts']
            
            return {
                'task_name': task_name,
                'success_rate': metrics['successes'] / total if total else 0,
                'failure_rate': metrics['failures'] / total if total else 0,
                'avg_retry_count': statistics.mean(metrics['retries']) if metrics['retries'] else 0,
                'total_attempts': total
            }
        
        summary = {}
        for task, metrics in self.task_metrics.items():
            total = metrics['attempts']
            summary[task] = {
                'success_rate': metrics['successes'] / total if total else 0,
                'attempts': total
            }
        return summary


class ToolInvocationEvaluator:
    """Tool & API Invocation Evaluation system."""
    
    def __init__(self):
        self.tool_calls: List[Dict[str, Any]] = []
        self.tool_metrics: Dict[str, Dict[str, Any]] = defaultdict(
            lambda: {
                'calls': 0, 'successful': 0, 'failed': 0,
                'invalid_params': 0, 'latencies': []
            }
        )
    
    def record_tool_call(self,
                        tool_name: str,
                        parameters: Dict[str, Any],
                        success: bool,
                        valid_parameters: bool = True,
                        latency_ms: float = None,
                        error: str = None,
                        metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        """Record tool invocation."""
        call = {
            'id': str(uuid.uuid4()),
            'tool_name': tool_name,
            'parameters': parameters,
            'success': success,
            'valid_parameters': valid_parameters,
            'latency_ms': latency_ms,
            'error': error,
            'metadata': metadata or {},
            'timestamp': time.time()
        }
        
        self.tool_calls.append(call)
        
        metrics = self.tool_metrics[tool_name]
        metrics['calls'] += 1
        if success:
            metrics['successful'] += 1
        else:
            metrics['failed'] += 1
        if not valid_parameters:
            metrics['invalid_params'] += 1
        if latency_ms:
            metrics['latencies'].append(latency_ms)
        
        return call
    
    def get_tool_metrics(self, tool_name: str = None) -> Dict[str, Any]:
        """Get tool usage metrics."""
        if tool_name:
            if tool_name not in self.tool_metrics:
                return {'error': 'Tool not found'}
            
            metrics = self.tool_metrics[tool_name]
            total = metrics['calls']
            
            return {
                'tool_name': tool_name,
                'success_rate': metrics['successful'] / total if total else 0,
                'failure_rate': metrics['failed'] / total if total else 0,
                'invalid_param_rate': metrics['invalid_params'] / total if total else 0,
                'avg_latency_ms': statistics.mean(metrics['latencies']) if metrics['latencies'] else 0,
                'total_calls': total
            }
        
        summary = {}
        for tool, metrics in self.tool_metrics.items():
            total = metrics['calls']
            summary[tool] = {
                'success_rate': metrics['successful'] / total if total else 0,
                'calls': total
            }
        return summary
    
    def detect_tool_call_patterns(self) -> Dict[str, Any]:
        """Detect common tool call patterns and issues."""
        if len(self.tool_calls) < 2:
            return {'error': 'Insufficient data'}
        
        patterns = {
            'repeated_failures': [],
            'slow_tools': [],
            'frequent_invalid_params': []
        }
        
        for tool, metrics in self.tool_metrics.items():
            total = metrics['calls']
            if total < 5:
                continue
            
            if metrics['failed'] / total > 0.3:
                patterns['repeated_failures'].append(tool)
            
            if metrics['latencies'] and statistics.mean(metrics['latencies']) > 5000:
                patterns['slow_tools'].append(tool)
            
            if metrics['invalid_params'] / total > 0.2:
                patterns['frequent_invalid_params'].append(tool)
        
        return patterns


__all__ = ['TaskEvaluator', 'ToolInvocationEvaluator']
